from __future__ import annotations

import enum
from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict, colors, datatypes, listdelegators


SceneLayerStr = Literal["item", "background", "foreground", "all"]

SCENE_LAYER: bidict[SceneLayerStr, widgets.QGraphicsScene.SceneLayer] = bidict(
    item=widgets.QGraphicsScene.SceneLayer.ItemLayer,
    background=widgets.QGraphicsScene.SceneLayer.BackgroundLayer,
    foreground=widgets.QGraphicsScene.SceneLayer.ForegroundLayer,
    all=widgets.QGraphicsScene.SceneLayer.AllLayers,
)

ItemIndexMethodStr = Literal["bsp_tree", "none"]

ITEM_INDEX_METHOD: bidict[ItemIndexMethodStr, widgets.QGraphicsScene.ItemIndexMethod] = (
    bidict(
        bsp_tree=widgets.QGraphicsScene.ItemIndexMethod.BspTreeIndex,
        none=widgets.QGraphicsScene.ItemIndexMethod.NoIndex,
    )
)


class GraphicsScene(core.ObjectMixin, widgets.QGraphicsScene):
    """Surface for managing a large number of 2D graphical items."""

    class GridType(enum.IntEnum):
        """Grid type for background."""

        NoGrid = 0
        DotGrid = 1
        LineGrid = 2

    def __init__(self, parent=None):
        super().__init__(parent)
        self._grid_mode = self.GridType.LineGrid
        self._grid_size = 50
        self._pen_width = 0.65
        self._grid_color = self.get_palette().get_color("text")
        self._bg_color = self.get_palette().get_color("window")
        # self.setBackgroundBrush(self._bg_color)

    def __repr__(self):
        cls_name = str(self.__class__.__name__)
        return f'<{cls_name}("{self.viewer()}") object at {hex(id(self))}>'

    def __getitem__(self, index: int) -> widgets.QGraphicsItem:
        return self.items()[index]

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"itemIndexMethod": ITEM_INDEX_METHOD}
        return maps

    def get_palette(self) -> gui.Palette:
        return gui.Palette(self.palette())

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_background_brush(self) -> gui.Brush:
        return gui.Brush(self.backgroundBrush())

    def get_foreground_brush(self) -> gui.Brush:
        return gui.Brush(self.foregroundBrush())

    def add(self, item) -> widgets.QGraphicsItem:
        match item:
            case widgets.QGraphicsItem():
                self.addItem(item)
                return item
            case gui.QPixmap():
                return self.add_pixmap(item)
            case gui.QPainterPath():
                return self.add_path(item)
            case gui.QPolygonF():
                return self.add_polygon(item)
            case core.QRectF():
                return self.add_rect(item)
            case core.QLine():
                return self.add_line(item)
            case str():
                return self.add_text(item)
            case widgets.QWidget():
                return self.add_widget(item)
            case _:
                raise TypeError(item)

    def add_pixmap(self, pixmap: gui.QPixmap) -> widgets.GraphicsPixmapItem:
        g_item = widgets.GraphicsPixmapItem(pixmap)
        self.addItem(g_item)
        return g_item

    def add_polygon(
        self,
        polygon: gui.QPolygonF | gui.QPolygon,
        pen: gui.QPen | None = None,
        brush: gui.QBrush | None = None,
    ) -> widgets.GraphicsPolygonItem:
        if isinstance(polygon, gui.QPolygon):
            polygon = gui.PolygonF(polygon)
        g_item = widgets.GraphicsPolygonItem(polygon)
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_path(
        self,
        path: gui.QPainterPath,
        pen: gui.QPen | None = None,
        brush: gui.QBrush | None = None,
    ) -> widgets.GraphicsPathItem:
        g_item = widgets.GraphicsPathItem(path)
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_rect(
        self,
        rect: datatypes.RectType | datatypes.RectFType,
        pen: gui.QPen | None = None,
        brush: gui.QBrush | None = None,
    ) -> widgets.GraphicsRectItem:
        g_item = widgets.GraphicsRectItem(datatypes.to_rectf(rect))
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_line(
        self,
        line: core.QLineF | core.QLine | tuple[float, float, float, float],
        pen: gui.QPen | None = None,
    ) -> widgets.GraphicsLineItem:
        g_item = widgets.GraphicsLineItem(datatypes.to_linef(line))
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_ellipse(
        self,
        ellipse: datatypes.RectType | datatypes.RectFType,
        pen: gui.QPen | None = None,
        brush: gui.QBrush | None = None,
    ) -> widgets.GraphicsEllipseItem:
        g_item = widgets.GraphicsEllipseItem(datatypes.to_rectf(ellipse))
        if brush is not None:
            g_item.setBrush(brush)
        if pen is not None:
            g_item.setPen(pen)
        self.addItem(g_item)
        return g_item

    def add_text(
        self, text: str, font: gui.QFont | None = None
    ) -> widgets.GraphicsTextItem:
        g_item = widgets.GraphicsTextItem(text)
        if font is not None:
            g_item.setFont(font)
        self.addItem(g_item)
        return g_item

    def add_simple_text(
        self, text: str, font: gui.QFont | None = None
    ) -> widgets.GraphicsSimpleTextItem:
        g_item = widgets.GraphicsSimpleTextItem(text)
        if font is not None:
            g_item.setFont(font)
        self.addItem(g_item)
        return g_item

    def add_widget(self, widget: widgets.QWidget) -> widgets.GraphicsProxyWidget:
        g_item = widgets.GraphicsProxyWidget()
        g_item.setWidget(widget)
        self.addItem(g_item)
        return g_item

    def colliding_items(
        self,
        item: widgets.QGraphicsItem,
        mode: (
            constants.ItemSelectionModeStr | constants.ItemSelectionMode
        ) = "intersects_shape",
    ) -> listdelegators.ListDelegator[widgets.QGraphicsItem]:
        items = self.collidingItems(item, constants.ITEM_SELECTION_MODE[mode])
        return listdelegators.ListDelegator(items)

    def add_item_group(self, *items: widgets.QGraphicsItem) -> widgets.GraphicsItemGroup:
        group = widgets.GraphicsItemGroup()
        for item in items:
            group.addToGroup(item)
        return group

    def _draw_grid(
        self,
        painter: gui.QPainter,
        rect: core.QRectF,
        pen: gui.QPen,
        grid_size: int,
    ):
        left = int(rect.left())
        right = int(rect.right())
        top = int(rect.top())
        bottom = int(rect.bottom())

        first_left = left - (left % grid_size)
        first_top = top - (top % grid_size)

        lines = [
            core.QLineF(x, top, x, bottom) for x in range(first_left, right, grid_size)
        ]
        lines.extend([
            core.QLineF(left, y, right, y) for y in range(first_top, bottom, grid_size)
        ])

        painter.setPen(pen)
        painter.drawLines(lines)

    def _draw_dots(
        self,
        painter: gui.QPainter,
        rect: core.QRectF,
        pen: gui.QPen,
        grid_size: int,
    ):
        if (zoom := self._get_viewer_zoom()) < 0:
            grid_size *= int(abs(zoom) / 0.3 + 1)

        left = int(rect.left())
        right = int(rect.right())
        top = int(rect.top())
        bottom = int(rect.bottom())

        first_left = left - (left % grid_size)
        first_top = top - (top % grid_size)

        pen.setWidth(grid_size / 10)
        painter.setPen(pen)

        [
            painter.drawPoint(int(x), int(y))
            for x in range(first_left, right, grid_size)
            for y in range(first_top, bottom, grid_size)
        ]

    def drawBackground(self, painter: gui.QPainter, rect: core.QRect):
        super().drawBackground(painter, rect)

        painter.save()
        painter.setRenderHint(gui.QPainter.RenderHint.Antialiasing, False)
        painter.setBrush(self.backgroundBrush())

        if self._grid_mode == self.GridType.DotGrid:
            pen = gui.QPen(self.grid_color, self._pen_width)
            self._draw_dots(painter, rect, pen, self._grid_size)

        elif self._grid_mode == self.GridType.LineGrid:
            zoom = self._get_viewer_zoom()
            if zoom > -0.5:
                pen = gui.QPen(self.grid_color, self._pen_width)
                self._draw_grid(painter, rect, pen, self._grid_size)
            color = self._bg_color.darker(150)
            if zoom < -0.0:
                color = color.darker(100 - int(zoom * 110))
            pen = gui.QPen(color, self._pen_width)
            self._draw_grid(painter, rect, pen, self._grid_size * 8)

        painter.restore()

    def _get_viewer_zoom(self):
        viewer = self.viewer()
        if viewer is None:
            return 1.0
        transform = viewer.transform()
        cur_scale = (transform.m11(), transform.m22())
        return float(f"{cur_scale[0] - 1.0:0.2f}")

    # def mousePressEvent(self, event):
    #     selected = self.viewer().selectedItems()
    #     if viewer := self.viewer():
    #         viewer.sceneMousePressEvent(event)
    #     super().mousePressEvent(event)
    #     keep_selection = any(
    #         [
    #             event.button() == core.Qt.MiddleButton,
    #             event.button() == core.Qt.RightButton,
    #             event.modifiers() == core.Qt.AltModifier,
    #         ]
    #     )
    #     if keep_selection:
    #         for node in selected:
    #             node.setSelected(True)

    # def mouseMoveEvent(self, event):
    #     if viewer := self.viewer():
    #         viewer.sceneMouseMoveEvent(event)
    #     super().mouseMoveEvent(event)

    # def mouseReleaseEvent(self, event):
    #     if viewer := self.viewer():
    #         viewer.sceneMouseReleaseEvent(event)
    #     super().mouseReleaseEvent(event)

    def viewer(self):
        return self.views()[0] if self.views() else None

    def get_grid_mode(self) -> GridType:
        return self._grid_mode

    def set_grid_mode(self, mode: GridType | None = None):
        # alternative?
        # brush = gui.Brush()
        # brush.set_style("cross")
        # scene.setBackgroundBrush(brush)
        if mode is None:
            mode = self.GridType.NoGrid
        self._grid_mode = mode

    def get_grid_color(self) -> gui.Color:
        return self._grid_color

    def set_grid_color(self, color: datatypes.ColorType):
        self._grid_color = colors.get_color(color)

    def get_background_color(self) -> gui.Color:
        return self._bg_color

    def set_background_color(self, color: datatypes.ColorType):
        self._bg_color = colors.get_color(color)
        self.setBackgroundBrush(self._bg_color)

    def set_item_index_method(
        self, method: ItemIndexMethodStr | widgets.QGraphicsScene.ItemIndexMethod
    ):
        """Set item index method.

        Args:
            method: item index method to use
        """
        self.setItemIndexMethod(ITEM_INDEX_METHOD.get_enum_value(method))

    def get_item_index_method(self) -> ItemIndexMethodStr:
        """Return item index method.

        Returns:
            item index method
        """
        return ITEM_INDEX_METHOD.inverse[self.itemIndexMethod()]

    bg_color = core.Property(
        gui.QColor,
        get_background_color,
        set_background_color,
        doc="Color for the scene background",
    )
    grid_color = core.Property(
        gui.QColor,
        get_grid_color,
        set_grid_color,
        doc="Color for the grid",
    )
    grid_mode = core.Property(
        int,
        get_grid_mode,
        set_grid_mode,
        doc="Grid mode",
    )


if __name__ == "__main__":
    app = widgets.app()
    scene = GraphicsScene()
    scene.add_line(core.Line(0, 0, 10, 10))
    view = widgets.GraphicsView(scene)
    view.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, gui


class PainterPathStroker(gui.QPainterPathStroker):
    def set_cap_style(self, style: constants.CapStyleStr | constants.PenCapStyle):
        """Set cap style to use.

        Args:
            style: cap style to use
        """
        self.setCapStyle(constants.CAP_STYLE.get_enum_value(style))

    def get_cap_style(self) -> constants.CapStyleStr | constants.PenCapStyle:
        """Return current cap style.

        Returns:
            cap style
        """
        return constants.CAP_STYLE.inverse[self.capStyle()]

    def set_join_style(self, style: constants.JoinStyleStr | constants.PenJoinStyle):
        """Set join style to use.

        Args:
            style: join style to use
        """
        self.setJoinStyle(constants.JOIN_STYLE.get_enum_value(style))

    def get_join_style(self) -> constants.JoinStyleStr:
        """Return current join style.

        Returns:
            join style
        """
        return constants.JOIN_STYLE.inverse[self.joinStyle()]

    def create_stroke(self, path: gui.QPainterPath) -> gui.PainterPath:
        return gui.PainterPath(self.createStroke(path))


if __name__ == "__main__":
    p = PainterPathStroker()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict


StackingModeStr = Literal["one", "all"]

STACKING_MODE: bidict[StackingModeStr, widgets.QStackedLayout.StackingMode] = bidict(
    one=widgets.QStackedLayout.StackingMode.StackOne,
    all=widgets.QStackedLayout.StackingMode.StackAll,
)


class StackedLayout(widgets.LayoutMixin, widgets.QStackedLayout):
    """Layout containing stack of widgets where only one widget is visible at a time."""

    ID = "stacked"

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"stackingMode": STACKING_MODE}
        return maps

    def set_stacking_mode(self, mode: StackingModeStr):
        self.setStackingMode(STACKING_MODE[mode])

    def get_stacking_mode(self) -> StackingModeStr:
        return STACKING_MODE.inverse[self.stackingMode()]

    def __add__(self, other: widgets.QWidget | widgets.QLayout):
        self.add(other)
        return self

    def set_current_widget(self, widget: widgets.QWidget):
        self.setCurrentWidget(widget)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    layout = StackedLayout()
    widget = widgets.Widget()
    widget2 = widgets.RadioButton("Test")
    widget3 = widgets.RadioButton("Test 2")
    layout += widget2
    layout += widget3
    widget.set_layout(layout)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtScxml


class ScxmlInvokableServiceFactory(
    core.ObjectMixin, QtScxml.QScxmlInvokableServiceFactory
):
    """Creates invokable service instances."""



from __future__ import annotations

import os
from typing import Self

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import colors, datatypes, get_repr


class Label(widgets.FrameMixin, widgets.QLabel):
    """Text or image display."""

    elision_changed = core.Signal(bool)
    clicked = core.Signal()

    def __init__(self, *args, **kwargs):
        self._elide_mode = constants.TextElideMode.ElideNone
        super().__init__(*args, **kwargs)
        self.openExternalLinks()
        self._is_elided = False

    def mouseReleaseEvent(self, ev: gui.QMouseEvent):
        if ev.button() == constants.MouseButton.LeftButton:
            self.clicked.emit()
        return super().mouseReleaseEvent(ev)

    def set_elide_mode(self, mode: constants.TextElideModeStr):
        self._elide_mode = constants.TEXT_ELIDE_MODE[mode]
        self.update()

    def get_elide_mode(self) -> constants.TextElideModeStr:
        return constants.TEXT_ELIDE_MODE.inverse[self._elide_mode]

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"textFormat": constants.TEXT_FORMAT, "alignment": constants.ALIGNMENTS}
        return maps

    def __repr__(self):
        return get_repr(self, self.text())

    # # adapted from https://forum.qt.io/topic/24530/solved-shortening-a-label/3
    # def minimumSizeHint(self):
    #     if self._elide_mode != constants.TextElideMode.ElideNone:
    #         # TODO: tweak sizeHint
    #         # -> text should expand if user increases window size,
    #         #    but don't automatically adapt window size to label width on UI update!
    #         #    (somehow calculate minimumSizeHint + sizeHint with font metrics???)
    #         fm = self.fontMetrics()
    #         size = core.QSize(fm.width("..."), fm.height())
    #         return size
    #     else:
    #         size = self.minimumSizeHint()
    #         return core.QSize(size.width() + 13, size.height())

    # # adapted from https://www.mimec.org/blog/status-bar-and-elided-label
    # def paintEvent(self, event):
    #     with gui.Painter(self) as painter:
    #         self.drawFrame(painter)

    #         rect = self.contentsRect()
    #         rect.adjust(self.margin(), self.margin(), -self.margin(), -self.margin())

    #         elided_text = painter.fontMetrics().elidedText(
    #             self.text(), self._elide_mode, rect.width()
    #         )

    #         style_option = widgets.QStyleOption()
    #         style_option.initFrom(self)

    #         self.style().drawItemText(
    #             painter,
    #             rect,
    #             self.alignment(),
    #             style_option.palette,
    #             self.isEnabled(),
    #             elided_text,
    #             self.foregroundRole(),
    #         )

    def paintEvent(self, event):
        if self._elide_mode == constants.TextElideMode.ElideNone:
            super().paintEvent(event)
            return
        did_elide = False

        with gui.Painter(self) as painter:
            font_metrics = painter.fontMetrics()
            text_lines = self.text().split("\n")
            text_width = font_metrics.horizontalAdvance(self.text())
            line_spacing = font_metrics.lineSpacing()

            # layout phase
            text_layout = gui.TextLayout(self.text(), painter.font())
            current_y = 0
            with text_layout.process_layout():
                for line in text_lines:
                    text_width = font_metrics.horizontalAdvance(line)
                    # if self.height() >= next_line_y + line_spacing:
                    #     line.draw(painter, core.PointF(0, y))
                    #     y = next_line_y
                    # else:
                    #     last_line = self._text[line.textStart() :]
                    #     elided_line = metrics.elided_text(
                    #         last_line, "right", self.width()
                    #     )
                    #     painter.drawText(0, y + metrics.ascent(), elided_line)
                    #     line = layout.createLine()
                    #     did_elide = line.isValid()
                    #     break
                    if text_width >= self.width():
                        elided_line = font_metrics.elidedText(
                            line, self._elide_mode, self.width()
                        )
                        painter.drawText(
                            core.QRect(0, current_y, self.width(), self.height()),
                            int(self.alignment()),
                            elided_line,
                        )
                        did_elide = True
                    else:
                        painter.drawText(
                            core.QRect(0, current_y, self.width(), self.height()),
                            int(self.alignment()),
                            line,
                        )
                    current_y += line_spacing

            if did_elide != self._is_elided:
                self._is_elided = did_elide
                self.elision_changed.emit(did_elide)

    def allow_links(self) -> Label:
        # self.setText("<a href=\"http://example.com/\">Click Here!</a>")
        self.setTextFormat(constants.TextFormat.RichText)
        self.setTextInteractionFlags(constants.TextInteractionFlag.TextBrowserInteraction)
        self.setOpenExternalLinks(True)
        return self

    def set_alignment(
        self,
        horizontal: constants.HorizontalAlignmentStr | None = None,
        vertical: constants.VerticalAlignmentStr | None = None,
    ):
        """Set the alignment of the label's contents."""
        match horizontal, vertical:
            case None, None:
                return self
            case None, _:
                flag = constants.V_ALIGNMENT[vertical]
            case _, None:
                flag = constants.H_ALIGNMENT[horizontal]
            case _, _:
                flag = constants.V_ALIGNMENT[vertical] | constants.H_ALIGNMENT[horizontal]
        self.setAlignment(flag)
        return self

    def get_horizontal_alignment(self) -> constants.HorizontalAlignmentStr:
        align = self.alignment()
        if align & constants.ALIGN_RIGHT:  # type: ignore
            return "right"
        if align & constants.ALIGN_H_CENTER:  # type: ignore
            return "center"
        if align & constants.ALIGN_JUSTIFY:  # type: ignore
            return "justify"
        return "left"

    def get_vertical_alignment(self) -> constants.VerticalAlignmentStr:
        align = self.alignment()
        if align & constants.ALIGN_TOP:  # type: ignore
            return "top"
        if align & constants.ALIGN_BOTTOM:  # type: ignore
            return "bottom"
        if align & constants.ALIGN_BASELINE:  # type: ignore
            return "baseline"
        return "center"

    def set_indent(self, indent: int) -> Label:
        """Set the label's text indent in pixels."""
        self.setIndent(indent)
        return self

    def set_text_format(
        self, text_format: constants.TextFormatStr | constants.TextFormat
    ) -> Label:
        """Set the text format.

        Args:
            text_format: text format to use
        """
        self.setTextFormat(constants.TEXT_FORMAT.get_enum_value(text_format))
        return self

    def get_text_format(self) -> constants.TextFormatStr:
        """Return current text format.

        Returns:
            text format
        """
        return constants.TEXT_FORMAT.inverse[self.textFormat()]

    def set_text_interaction(self, *types: constants.TextInteractionStr) -> Label:
        """Set the text interaction mode.

        Args:
            types: text interaction mode to use
        """
        flags = constants.TEXT_INTERACTION.merge_flags(types)
        self.setTextInteractionFlags(flags)
        return self

    def get_text_interaction(self) -> list[constants.TextInteractionStr]:
        """Return current text interaction mode.

        Returns:
            list of text interaction modes
        """
        return constants.TEXT_INTERACTION.get_list(self.textInteractionFlags())

    def set_text(self, text: str) -> Label:
        """Set the label's text."""
        self.setText(text)
        return self

    def set_bold(self, bold: bool = True) -> Label:
        font = self.font()
        font.setBold(bold)
        self.setFont(font)
        return self

    def set_italic(self, italic: bool = True) -> Label:
        font = self.font()
        font.setItalic(italic)
        self.setFont(font)
        return self

    def set_point_size(self, size: int) -> Label:
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)
        return self

    def set_weight(self, weight: gui.font.WeightStr | gui.QFont.Weight) -> Label:
        """Set the font weight.

        Args:
            weight: font weight
        """
        font = self.font()
        font.setWeight(gui.font.WEIGHT.get_enum_value(weight))
        self.setFont(font)
        return self

    def set_color(self, color: datatypes.ColorType) -> Label:
        with self.edit_stylesheet() as ss:
            if color is None:
                ss.color.setValue("")
            else:
                color = colors.get_color(color)
                ss.color.setValue(color.name())
        return self

    def set_image(self, path: datatypes.PathType, width: int = 300) -> Label:
        self.setScaledContents(True)
        self.set_alignment(horizontal="center")
        self.setText(
            "<html><head/><body><p>"
            f"<img src={os.fspath(path)!r} width={str(width)!r}/>"
            "</p></body></html>"
        )
        return self

    @classmethod
    def image_from_path(
        cls, path: datatypes.PathType, parent: widgets.QWidget | None = None
    ) -> Self:
        pixmap = gui.Pixmap.from_file(path)
        label = cls(parent=parent)
        label.setPixmap(pixmap)
        label.resize(pixmap.width(), pixmap.height())
        return label

    elideMode = core.Property(  # noqa: N815
        str,
        get_elide_mode,
        set_elide_mode,
        doc="Text Elide style",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = Label("http://www.test.fsdfsdfsfdsfsfdsfde\n" * 20, elide_mode="right")
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core
from prettyqt.qt import QtWidgets
from prettyqt.utils import datatypes


if TYPE_CHECKING:
    from collections.abc import Callable


class SlideAnimation(core.PropertyAnimation):
    ID = "slide"

    def __init__(
        self,
        parent: QtWidgets.QWidget,
        duration: int = 1000,
        start: tuple[int, int] | core.Point = (0, 0),
        end: tuple[int, int] | core.Point = (0, 0),
        easing: core.easingcurve.TypeStr = "in_out_sine",
    ):
        super().__init__(parent)
        self.set_easing(easing)
        self.set_start_value(start)
        self.set_end_value(end)
        self.setDuration(duration)

    def set_start_value(self, point: datatypes.PointType):
        offset = datatypes.to_point(point)
        pos = self.parent().geometry().topLeft()
        self.setStartValue(pos + offset)

    def set_end_value(self, point: datatypes.PointType):
        offset = datatypes.to_point(point)
        pos = self.parent().geometry().topLeft()
        self.setEndValue(pos + offset)

    def apply_to(self, obj: QtWidgets.QWidget | Callable):
        if isinstance(obj, QtWidgets.QWidget):
            obj = obj.pos
        super().apply_to(obj)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    btn = widgets.PushButton("ts")
    btn.show()
    val = SlideAnimation(parent=btn)
    val.set_end_value((0, 100))
    val.apply_to(btn)
    val.start()
    app.exec()



from __future__ import annotations

import os
import pathlib
from typing import Literal

from prettyqt import core, qml
from prettyqt.utils import bidict, datatypes


ObjectOwnershipStr = Literal["cpp", "javascript"]


OBJECT_OWNERSHIP: bidict[ObjectOwnershipStr, qml.QQmlEngine.ObjectOwnership] = bidict(
    cpp=qml.QQmlEngine.ObjectOwnership.CppOwnership,
    javascript=qml.QQmlEngine.ObjectOwnership.JavaScriptOwnership,
)


class QmlEngineMixin(qml.JSEngineMixin):
    def set_object_ownership(
        self,
        obj: core.QObject,
        mode: ObjectOwnershipStr | qml.QQmlEngine.ObjectOwnership,
    ):
        """Set the object ownership.

        Args:
            obj: object to set ownership for
            mode: object ownership to use
        """
        self.setObjectOwnership(obj, OBJECT_OWNERSHIP.get_enum_value(mode))

    def get_object_ownership(self, obj: core.QObject) -> ObjectOwnershipStr:
        """Return object ownership.

        Returns:
            object ownership
        """
        return OBJECT_OWNERSHIP.inverse[self.objectOwnership(obj)]

    def add_import_path(self, path: datatypes.PathType):
        self.addImportPath(os.fspath(path))

    def add_plugin_path(self, path: datatypes.PathType):
        self.addPluginPath(os.fspath(path))

    def get_plugin_paths(self) -> list[pathlib.Path]:
        return [pathlib.Path(p) for p in self.pluginPathList()]

    def get_import_paths(self) -> list[pathlib.Path]:
        return [pathlib.Path(p) for p in self.importPathList()]

    def set_base_url(self, url: str | core.QUrl):
        if isinstance(url, str):
            url = core.QUrl(url)
        self.setBaseUrl(url)

    def get_base_url(self) -> core.Url:
        return core.Url(self.baseUrl())

    def set_offline_storage_path(self, path: datatypes.PathType):
        self.setOfflineStoragePath(os.fspath(path))


class QmlEngine(QmlEngineMixin, qml.QQmlEngine):
    """Environment for instantiating QML components."""



from __future__ import annotations

from typing import Literal

from prettyqt import core, widgets
from prettyqt.utils import bidict, datatypes


ViewModeStr = Literal["list", "icon"]

VIEW_MODE: bidict[ViewModeStr, widgets.QListView.ViewMode] = bidict(
    list=widgets.QListView.ViewMode.ListMode, icon=widgets.QListView.ViewMode.IconMode
)

MovementStr = Literal["static", "free", "snap"]

MOVEMENT: bidict[MovementStr, widgets.QListView.Movement] = bidict(
    static=widgets.QListView.Movement.Static,
    free=widgets.QListView.Movement.Free,
    snap=widgets.QListView.Movement.Snap,
)

FlowStr = Literal["left_to_right", "top_to_bottom"]

FLOW: bidict[FlowStr, widgets.QListView.Flow] = bidict(
    left_to_right=widgets.QListView.Flow.LeftToRight,
    top_to_bottom=widgets.QListView.Flow.TopToBottom,
)

LayoutModeStr = Literal["single_pass", "batched"]

LAYOUT_MODE: bidict[LayoutModeStr, widgets.QListView.LayoutMode] = bidict(
    single_pass=widgets.QListView.LayoutMode.SinglePass,
    batched=widgets.QListView.LayoutMode.Batched,
)

ResizeModeStr = Literal["fixed", "adjust"]

RESIZE_MODE: bidict[ResizeModeStr, widgets.QListView.ResizeMode] = bidict(
    fixed=widgets.QListView.ResizeMode.Fixed,
    adjust=widgets.QListView.ResizeMode.Adjust,
)


class ListViewMixin(widgets.AbstractItemViewMixin):
    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "viewMode": VIEW_MODE,
            "resizeMode": RESIZE_MODE,
            "layoutMode": LAYOUT_MODE,
            "movement": MOVEMENT,
            "flow": FLOW,
        }
        return maps

    def set_view_mode(self, mode: ViewModeStr | widgets.QListView.ViewMode):
        """Set view mode.

        Args:
            mode: view mode to use
        """
        self.setViewMode(VIEW_MODE.get_enum_value(mode))

    def get_view_mode(self) -> ViewModeStr:
        """Return view mode.

        Returns:
            view mode
        """
        return VIEW_MODE.inverse[self.viewMode()]

    def set_resize_mode(self, mode: ResizeModeStr | widgets.QListView.ResizeMode):
        """Set resize mode.

        Args:
            mode: resize mode to use
        """
        self.setResizeMode(RESIZE_MODE.get_enum_value(mode))

    def get_resize_mode(self) -> ResizeModeStr:
        """Return resize mode.

        Returns:
            resize mode
        """
        return RESIZE_MODE.inverse[self.resizeMode()]

    def set_layout_mode(self, mode: LayoutModeStr | widgets.QListView.LayoutMode):
        """Set layout mode.

        Args:
            mode: layout mode to use
        """
        self.setLayoutMode(LAYOUT_MODE.get_enum_value(mode))

    def get_layout_mode(self) -> LayoutModeStr:
        """Return layout mode.

        Returns:
            layout mode
        """
        return LAYOUT_MODE.inverse[self.layoutMode()]

    def set_movement(self, mode: MovementStr | widgets.QListView.Movement):
        """Set movement mode.

        Args:
            mode: movement mode to use
        """
        self.setMovement(MOVEMENT.get_enum_value(mode))

    def get_movement(self) -> MovementStr:
        """Return movement mode.

        Returns:
            movement mode
        """
        return MOVEMENT.inverse[self.movement()]

    def set_flow(self, mode: FlowStr | widgets.QListView.Flow):
        """Set flow mode.

        Args:
            mode: flow mode to use
        """
        self.setFlow(FLOW.get_enum_value(mode))

    def get_flow(self) -> FlowStr:
        """Return flow mode.

        Returns:
            flow mode
        """
        return FLOW.inverse[self.flow()]

    def set_grid_size(self, size: datatypes.SizeType):
        self.setGridSize(datatypes.to_size(size))

    def get_grid_size(self) -> core.Size:
        return core.Size(self.gridSize())


class ListView(ListViewMixin, widgets.QListView):
    """List or icon view onto a model."""


if __name__ == "__main__":
    app = widgets.app()
    dlg = ListView()
    dlg.show()
    app.exec()



from __future__ import annotations

import contextlib

from prettyqt.qt import QtCore


class Mutex(QtCore.QMutex):
    """Access serialization between threads."""

    @contextlib.contextmanager
    def lock_mutex(self, timeout: int | None = None):
        if timeout is None:
            timeout = -1
        yield self.tryLock(timeout)
        self.unlock()



from __future__ import annotations

import sys

from prettyqt import core, gui


class IntegerValidator(gui.Validator):
    """Validator to check for integer values.

    In contrast to IntValidator provided by Qt, this validator is not bound to
    a restricted value range.
    """

    ID = "integer"

    def __init__(
        self,
        bottom: int | None = None,
        top: int | None = None,
        parent: core.QObject | None = None,
    ):
        super().__init__(parent)
        self._bottom = -sys.maxsize if bottom is None else bottom
        self._top = sys.maxsize if top is None else top

    def set_range(self, lower: int | None, upper: int | None):
        self.set_top(upper)
        self.set_bottom(lower)

    def get_range(self) -> tuple[int, int]:
        return (self._bottom, self._top)

    def set_bottom(self, bottom: int | None):
        self._bottom = -sys.maxsize if bottom is None else bottom

    def get_bottom(self) -> int:
        return self._bottom

    def set_top(self, top: int | None):
        self._top = sys.maxsize if top is None else top

    def get_top(self) -> int:
        return self._top

    def validate(self, text: str, pos: int):
        if not text.lstrip("-"):
            return self.State.Intermediate, text, len(text)
        if text.lstrip("-").isnumeric():
            left = float("-inf") if self._bottom is None else self._bottom
            right = float("inf") if self._top is None else self._top
            if left <= int(text) <= right:
                return self.State.Acceptable, text, len(text)
        return self.State.Invalid, text, len(text)

    bottom = core.Property(
        int,
        get_bottom,
        set_bottom,
        doc="Minimum value",
    )
    top = core.Property(
        int,
        get_top,
        set_top,
        doc="Maximum value",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    val = IntegerValidator()
    app = widgets.app()
    widget = widgets.LineEdit("343")
    widget.setValidator(val)
    widget.show()
    app.exec()



"""Tests for `datacook` package."""

from __future__ import annotations

import sys

import pytest

from prettyqt import constants, gui, widgets
from prettyqt.qt import QtCore, QtWidgets


# import pytest


@pytest.fixture
def tablewidget():
    widget = widgets.TableWidget()
    widget.setColumnCount(3)
    widget.insertRow(0)
    widget.setHorizontalHeaderLabels(["LIB", "CELL", "area"])
    item = widgets.TableWidgetItem("test")
    widget.setItem(0, 0, item)
    widget.setItem(1, 1, widgets.TableWidgetItem("test"))
    return widget


@pytest.fixture(scope="session")
def qapp_cls():
    return widgets.Application


@pytest.fixture(scope="session")
def qapp_args():
    return sys.argv


class QtTester:
    @staticmethod
    def send_keypress(widget: QtWidgets.QWidget, key):
        event = gui.QKeyEvent(
            QtCore.QEvent.Type.KeyPress,
            key,
            constants.KeyboardModifier(0),  # type: ignore
        )
        widgets.Application.sendEvent(widget, event)

    @staticmethod
    def send_mousepress(widget: QtWidgets.QWidget, key):
        event = gui.QMouseEvent(  # type: ignore
            QtCore.QEvent.Type.MouseButtonRelease,
            QtCore.QPointF(0, 0),
            QtCore.QPointF(0, 0),
            key,
            constants.MouseButton.NoButton,
            constants.KeyboardModifier(0),  # type: ignore
        )
        widgets.Application.sendEvent(widget, event)

    @staticmethod
    def send_mousemove(
        widget: QtWidgets.QWidget, target: QtCore.QPointF | None = None, delay: int = 0
    ):
        if target is None:
            target = QtCore.QPointF(0, 0)
        event = gui.QMouseEvent(  # type: ignore
            QtCore.QEvent.Type.MouseButtonRelease,
            target,
            QtCore.QPointF(0, 0),
            constants.MouseButton.NoButton,
            constants.MouseButton.NoButton,
            constants.KeyboardModifier(0),  # type: ignore
        )
        widgets.Application.sendEvent(widget, event)

    # @staticmethod
    # def test_model(model: QtCore.QAbstractItemModel, force_py: bool = False):
    #     tester = modeltest.ModelTester(model)
    #     tester.check(force_py=force_py)
    #     tester._cleanup()


@pytest.fixture
def qttester() -> type[QtTester]:
    return QtTester



"""Provides access to Bluetooth hardware."""

from __future__ import annotations

from prettyqt.qt.QtBluetooth import *  # noqa: F403

from .bluetoothdeviceinfo import BluetoothDeviceInfo
from .bluetoothuuid import BluetoothUuid
from .bluetoothaddress import BluetoothAddress
from .bluetoothhostinfo import BluetoothHostInfo
from .bluetoothlocaldevice import BluetoothLocalDevice
from .bluetoothserviceinfo import BluetoothServiceInfo
from .bluetoothsocket import BluetoothSocket
from .bluetoothserver import BluetoothServer
from .bluetoothdevicediscoveryagent import BluetoothDeviceDiscoveryAgent
from .bluetoothservicediscoveryagent import BluetoothServiceDiscoveryAgent
from prettyqt.qt import QtBluetooth

QT_MODULE = QtBluetooth


__all__ = [
    "BluetoothAddress",
    "BluetoothDeviceDiscoveryAgent",
    "BluetoothDeviceInfo",
    "BluetoothHostInfo",
    "BluetoothLocalDevice",
    "BluetoothServer",
    "BluetoothServiceDiscoveryAgent",
    "BluetoothServiceInfo",
    "BluetoothSocket",
    "BluetoothUuid",
]



from __future__ import annotations

from prettyqt import core, qthelp


class HelpFilterEngine(core.ObjectMixin):
    """Filtered view of the help contents."""

    def __init__(self, item: qthelp.QHelpFilterEngine):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_available_versions(self) -> list[core.VersionNumber]:
        return [core.VersionNumber(i) for i in self.availableVersions()]

    def get_filter_data(self, filter_name: str) -> qthelp.HelpFilterData:
        return qthelp.HelpFilterData(self.filterData(filter_name))


if __name__ == "__main__":
    core_engine = qthelp.HelpEngineCore("test")
    engine = HelpFilterEngine(core_engine)
    print(engine.get_filter_data("a"))



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


TypeStr = Literal["read", "write", "exception"]

TYPE: bidict[TypeStr, core.QSocketNotifier.Type] = bidict(
    read=core.QSocketNotifier.Type.Read,
    write=core.QSocketNotifier.Type.Write,
    exception=core.QSocketNotifier.Type.Exception,
)


class SocketNotifier(core.ObjectMixin, core.QSocketNotifier):
    """Support for monitoring activity on a file descriptor."""

    # def __repr__(self):
    #     return f"{type(self).__name__}({self.socket()}, {self.type()})"

    def get_type(self) -> TypeStr:
        """Return socket event type.

        Returns:
            socket event type
        """
        return TYPE.inverse[self.type()]


if __name__ == "__main__":
    app = core.app()
    import os

    read, write = os.pipe()
    notifier = SocketNotifier(read, core.QSocketNotifier.Type.Read)
    app.exec()



from __future__ import annotations

import logging

from prettyqt import core, eventfilters, widgets
from prettyqt.utils import bidict, datatypes, helpers


logger = logging.getLogger(__name__)


class WidgetEditor(widgets.Widget):
    value_changed = core.Signal(object)

    def __init__(self, qobject: core.QObject, connect: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.set_layout("form")
        self._qobject = qobject
        self._initial_prop_values = {}
        self.event_catcher = eventfilters.EventCatcher(
            include=["resize", "move"], parent=self._qobject
        )
        if connect:
            self.event_catcher.caught.connect(self._update_editors)
            self._qobject.installEventFilter(self.event_catcher)
        self.set_minimum_size(800, 1000)
        self._editors = bidict()
        self._metaobj = core.MetaObject(self._qobject.metaObject())
        for i, prop in enumerate(self._metaobj.get_properties()):
            value = prop.read(self._qobject)
            # typ = prop.get_meta_type().get_type()
            name = prop.name()
            logger.info(f"setting {name} editor to {value}")
            widget = datatypes.get_editor_for_value(value)
            if widget is None:
                logger.warning(f"No editor found for {value!r}")
                continue
            widget.set_value(value)
            widget.value_changed.connect(self._on_value_change)
            widget.setEnabled(prop.isWritable())
            label = helpers.to_snake(prop.name()).replace("_", " ")
            self.box[i, "left"] = widgets.Label(label)
            self.box[i, "right"] = widget
            self._initial_prop_values[name] = value
            self._editors[name] = widget
            if prop.hasNotifySignal() and connect:
                notify_signal = prop.get_notify_signal()
                signal_name = notify_signal.get_name()
                signal = self._qobject.__getattribute__(signal_name)
                signal.connect(self._update_editors)

    # @classmethod
    # def setup_example(cls):
    #     scrollarea = widgets.ScrollArea()
    #     return cls(scrollarea, connect=False)

    def _on_value_change(self):
        editor = self.sender()
        prop_name = self._editors.inverse[editor]
        prop = self._metaobj.get_property(prop_name)
        value = editor.get_value()
        value = datatypes.make_qtype(value)
        # logger.debug(f"setting {prop_name} to {value}")
        prop.write(self._qobject, value)
        # brute force
        if isinstance(self._qobject, widgets.QWidget):
            self._qobject.updateGeometry()
            self._qobject.repaint()
            if (parent := self._qobject.parentWidget()) is not None:
                parent.updateGeometry()
                parent.repaint()

    def _update_editors(self):
        for prop_name, editor in self._editors.items():
            value = self._qobject.property(prop_name)
            editor.set_value(value)


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.PlainTextEdit()
    w.show()
    editor = WidgetEditor(w)
    editor.show()
    with app.debug_mode():
        app.exec()



import ctypes
from ctypes.wintypes import HBITMAP, HICON, LPCWSTR, RECT  # , HWND
import enum
import logging
from typing import Literal, SupportsInt

import comtypes.client as cc
import comtypes.gen.TaskbarLib as tbl


# from comtypes.gen import _683BF642_E9CA_4124_BE43_67065B2FA653_0_1_0


logger = logging.getLogger(__name__)
cc.GetModule("./TaskbarLib.tlb")

taskbar = cc.CreateObject(
    "{56FDF344-FD6D-11d0-958A-006097C9A090}", interface=tbl.ITaskbarList3
)
dwmapi = ctypes.WinDLL("dwmapi")
taskbar.HrInit()


class State(enum.IntEnum):
    no_progress = 0
    indeterminate = 1
    normal = 2
    error = 4
    paused = 8


class TaskBarItem:
    def __init__(self, hwnd: SupportsInt | None = None):
        self.state = "normal"
        self.progress = 0
        self.tabs: list[int] = []
        if hwnd is None:
            hwnd = ctypes.windll.kernel32.GetConsoleWindow()
            taskbar.ActivateTab(hwnd)
        self.win_id = int(hwnd)

    def set_progress_state(
        self,
        value: Literal[
            "no_progress", "indeterminate", "normal", "error", "paused", "done"
        ],
    ):
        match value:
            case "no_progress":
                taskbar.SetProgressState(self.win_id, State.no_progress)
                self.state = "no_progress"
            case "indeterminate":
                taskbar.SetProgressState(self.win_id, State.indeterminate)
                self.state = "indeterminate"
            case "normal":
                taskbar.SetProgressState(self.win_id, State.normal)
                self.state = "normal"
            case "error":
                taskbar.SetProgressState(self.win_id, State.error)
                self.state = "error"
            case "paused":
                taskbar.SetProgressState(self.win_id, State.paused)
                self.state = "paused"
            case "done":
                ctypes.windll.user32.FlashWindow(self.win_id, True)
                self.set_progress_state("no_progress")
            case _:
                raise ValueError(value)

    def set_progress(self, value: int, total: int = 100):
        if 0 < value < total:
            self.set_progress_state("normal")
        else:
            self.set_progress_state("no_progress")
            self.set_progress_value(value, total)

    def set_progress_value(self, value: int, total: int = 100):
        if value > 100 or value < 0 or value > total:
            raise ValueError(value)
        result = taskbar.setProgressValue(self.win_id, value, total)
        return bool(result)

    def activate_tab(self):
        logger.info(f"Activate tab {self.win_id}")
        result = taskbar.ActivateTab(self.win_id)
        return bool(result)

    def set_thumbnail_clip(self, rect: tuple[int, int, int, int]) -> bool:
        logger.info(f"Set thumnail clip to {rect!r} for handle {self.win_id}")
        area = RECT(*rect)
        result = taskbar.SetThumbnailClip(self.win_id, area)
        return bool(result)

    def set_thumbnail_tooltip(self, tooltip: str) -> bool:
        logger.info(f"Set thumnail tooltip {tooltip!r} for handle {self.win_id}")
        result = taskbar.SetThumbnailTooltip(self.win_id, LPCWSTR(tooltip))
        return bool(result)

    def register_tab(self, tab_win_id: SupportsInt):
        tab_win_id = int(tab_win_id)
        logger.info(f"Register tab {tab_win_id} in handle {self.win_id}")
        set_window_attribute(tab_win_id, 7, True)
        set_window_attribute(tab_win_id, 10, True)
        result = taskbar.RegisterTab(tab_win_id, self.win_id)
        taskbar.SetTabOrder(tab_win_id, 0)
        taskbar.SetTabActive(tab_win_id, 0)
        self.tabs.append(tab_win_id)
        return bool(result)

    def unregister_tab(self, tab_win_id: SupportsInt):
        tab_win_id = int(tab_win_id)
        # if tab_win_id not in self.tabs:
        #     raise ValueError(f"no handle with {tab_win_id} registered.")
        logger.info(f"Register tab {tab_win_id}")
        result = taskbar.UnregisterTab(tab_win_id)
        self.tabs.remove(tab_win_id)
        return bool(result)

    def set_tab_active(self, tab_win_id: SupportsInt):
        tab_win_id = int(tab_win_id)
        logger.info(f"Set tab {tab_win_id} for handle {self.win_id} active")
        result = taskbar.SetTabActive(tab_win_id, self.win_id, 0)
        return bool(result)

    def set_overlay_icon(self, icon, description: str = ""):
        logger.info(f"Set overlay icon for {self.win_id}")

        # CreateIconFromResourceEx = windll.user32.CreateIconFromResourceEx
        # size_x, size_y = 32, 32
        # LR_DEFAULTCOLOR = 0

        # with open(png_path, "rb") as f:
        #     png = f.read()
        # hicon = CreateIconFromResourceEx(
        #     png, len(png), 1, 0x30000, size_x, size_y, LR_DEFAULTCOLOR
        # )

        overlay_icon = HICON(icon)
        accessible_name = LPCWSTR(description)
        result = taskbar.SetOverlayIcon(self.win_id, overlay_icon, accessible_name)
        return bool(result)

    def set_iconic_live_preview_bitmap(self, bitmap: HBITMAP):  # available via QImage
        logger.info(f"Set iconic live preview for {self.win_id}")
        dwmapi.DwmSetIconicLivePreviewBitmap(self.win_id, bitmap, 0)

    def set_iconic_thumbnail(self, thumbnail: HBITMAP):  # available via QImage
        logger.info(f"Set iconic thumbnail for {self.win_id}")
        # SM_CXSMICON = 49
        # SM_CYSMICON = 50
        # ico_x = GetSystemMetrics(SM_CXSMICON)
        # ico_y = GetSystemMetrics(SM_CYSMICON)
        # icon = encode_for_locale(self._icon)
        # hicon = LoadImageA(0, icon, IMAGE_ICON, ico_x, ico_y, LR_LOADFROMFILE)
        # dwmapi.DwmSetIconicThumbnail(self.win_id, HBITMAP(thumbnail), 0)


def set_window_attribute(win_id, key, value):  # available via QImage
    logger.info(f"Set iconic thumbnail for {win_id}")
    dwmapi.DwmSetWindowAttribute(
        win_id,
        key,  # DWMWA_FORCE_ICONIC_REPRESENTATION = 7
        # DWMWA_HAS_ICONIC_BITMAP = 10
        ctypes.byref(ctypes.c_bool(value)),
        ctypes.sizeof(ctypes.c_bool(value)),
    )


if __name__ == "__main__":
    import time

    prog = TaskBarItem()
    prog.set_iconic_live_preview_bitmap(HBITMAP())
    print("state indeterminate")
    prog.set_progress_state("indeterminate")
    prog.set_thumbnail_tooltip("jkfdsjfkjk")
    prog.set_thumbnail_clip((0, 20, 50, 80))
    time.sleep(5)
    # print("state normal, set progress")
    # prog.set_progress_state("normal")

    # for i in range(100):
    #     prog.set_progress_value(i)
    #     time.sleep(0.05)
    # print("state paused, set progress")
    # prog.set_progress_value(0)

    # prog.set_progress_state("paused")

    # for i in range(100):
    #     prog.set_progress_value(i)
    #     time.sleep(0.05)

    # print("state error, set progress")
    # prog.set_progress_value(0)
    # prog.set_progress_state("error")

    # for i in range(100):
    #     prog.set_progress_value(i)
    #     time.sleep(0.05)

    # print("state done.")
    # prog.set_progress_value(0)

    # prog.set_progress_state("done")
    # time.sleep(3)



from __future__ import annotations

import pathlib

from prettyqt.qt import QtScxml


class ScxmlCompiler(QtScxml.QScxmlCompiler):
    """Compiler for SCXML files."""

    def get_file_name(self) -> pathlib.Path:
        return pathlib.Path(self.fileName())


if __name__ == "__main__":
    from prettyqt.qt import QtCore

    reader = QtCore.QXmlStreamReader()
    compiler = ScxmlCompiler(reader)



from __future__ import annotations

from prettyqt.qt import QtCore


class ByteArray(QtCore.QByteArray):
    """Array of bytes."""

    def __reduce__(self):
        return type(self), (bytes(self),)



"""Tests for `prettyqt` package."""

import pytest

from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


statemachine = pytest.importorskip("statemachine")


def test_historystate():
    state = statemachine.HistoryState()
    state.set_history_type("deep")
    assert state.get_history_type() == "deep"
    with pytest.raises(InvalidParamError):
        state.set_history_type("test")


def test_signaltransition():
    trans = statemachine.SignalTransition()
    trans.set_transition_type("parallel")
    assert trans.get_transition_type() == "parallel"
    with pytest.raises(InvalidParamError):
        trans.set_transition_type("test")


def test_state():
    state = statemachine.State()
    state.set_child_mode("parallel")
    assert state.get_child_mode() == "parallel"
    with pytest.raises(InvalidParamError):
        state.set_child_mode("test")


def test_statemachine():
    machine = statemachine.StateMachine()
    state = statemachine.State()
    machine += state
    assert machine.get_error() == "none"
    machine.set_global_restore_policy("dont_restore")
    assert machine.get_global_restore_policy() == "dont_restore"
    with pytest.raises(InvalidParamError):
        machine.set_global_restore_policy("test")
    event = QtCore.QEvent(QtCore.QEvent.ActionAdded)
    machine.post_event(event)
    with pytest.raises(InvalidParamError):
        machine.post_event(event, "test")



# type: ignore
# not available in PySide2

from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict, get_repr


LoadHintStr = Literal[
    "resolve_all_symbols",
    "export_external_symbols",
    "load_archive_member",
    "prevent_unload",
    "deep_bind",
]

LOAD_HINTS: bidict[LoadHintStr, core.QLibrary.LoadHint] = bidict(
    resolve_all_symbols=core.QLibrary.LoadHint.ResolveAllSymbolsHint,
    export_external_symbols=core.QLibrary.LoadHint.ExportExternalSymbolsHint,
    load_archive_member=core.QLibrary.LoadHint.LoadArchiveMemberHint,
    prevent_unload=core.QLibrary.LoadHint.PreventUnloadHint,
    deep_bind=core.QLibrary.LoadHint.DeepBindHint,
)


class Library(core.ObjectMixin, core.QLibrary):
    """Loads shared libraries at runtime."""

    def __bool__(self):
        return self.isLoaded()

    def __repr__(self):
        return get_repr(self, self.fileName())

    def get_load_hints(self) -> list[LoadHintStr]:
        return LOAD_HINTS.get_list(self.loadHints())

    def set_load_hints(self, **kwargs):
        flag = core.QLibrary.LoadHint(0)
        for k, v in kwargs.items():
            if v is True:
                flag |= LOAD_HINTS[k]
        self.setLoadHints(flag)



from __future__ import annotations

from prettyqt import core
from prettyqt.utils import get_repr


class PauseAnimation(core.AbstractAnimationMixin, core.QPauseAnimation):
    """Pause for QSequentialAnimationGroup."""

    def __repr__(self):
        return get_repr(self, self.duration())



from __future__ import annotations

from typing import Literal

from prettyqt import multimedia
from prettyqt.utils import bidict


ModeStr = Literal["null", "input", "output"]

MODE: bidict[ModeStr, multimedia.QAudioDevice.Mode] = bidict(
    null=multimedia.QAudioDevice.Mode.Null,
    input=multimedia.QAudioDevice.Mode.Input,
    output=multimedia.QAudioDevice.Mode.Output,
)


class AudioDevice(multimedia.QAudioDevice):
    def get_mode(self) -> ModeStr:
        return MODE.inverse[self.mode()]

    def get_supported_sample_formats(
        self,
    ) -> list[multimedia.audioformat.SampleFormatStr]:
        return multimedia.audioformat.SAMPLE_FORMAT.get_list(
            self.supportedSampleFormats()
        )

    def get_channel_config(self) -> multimedia.audioformat.ChannelConfigStr:
        return multimedia.audioformat.CHANNEL_CONFIG.inverse[self.channelConfiguration()]

    def get_id(self) -> str:
        return self.id().data().decode()


if __name__ == "__main__":
    fmt = AudioDevice()
    print(fmt.get_channel_config())



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import get_repr


class GraphicsSimpleTextItem(
    widgets.AbstractGraphicsShapeItemMixin, widgets.QGraphicsSimpleTextItem
):
    """Simple text path item that you can add to a QGraphicsScene."""

    def __repr__(self):
        return get_repr(self, self.text())



from __future__ import annotations

from typing import Literal

from prettyqt import gui, widgets
from prettyqt.qt import QtCore
from prettyqt.utils import bidict, colors, datatypes


OptionStr = Literal["show_alpha", "no_buttons", "no_native"]

OPTIONS: bidict[OptionStr, widgets.QColorDialog.ColorDialogOption] = bidict(
    show_alpha=widgets.QColorDialog.ColorDialogOption.ShowAlphaChannel,
    no_buttons=widgets.QColorDialog.ColorDialogOption.NoButtons,
    no_native=widgets.QColorDialog.ColorDialogOption.DontUseNativeDialog,
)


class ColorDialog(widgets.DialogMixin, widgets.QColorDialog):
    """Dialog widget for specifying colors."""

    @classmethod
    def get_color(
        cls,
        preset: datatypes.ColorType = None,
        allow_alpha: bool = False,
        parent: widgets.QWidget | None = None,
    ) -> gui.Color:
        preset = colors.get_color(preset)
        kwargs = (
            dict(options=cls.ColorDialogOption.ShowAlphaChannel) if allow_alpha else {}
        )
        color = cls.getColor(preset, parent, **kwargs)  # type: ignore
        return gui.Color(color)

    def current_color(self) -> gui.Color:
        return gui.Color(self.currentColor())

    def get_qcolorshower(self) -> widgets.QWidget:
        return next(
            a
            for a in self.children()
            if hasattr(a, "metaObject")
            and a.metaObject().className() == "QtPrivate::QColorShower"
        )

    def get_qcolorshowlabel(self) -> widgets.QFrame:
        qcs = self.get_qcolorshower()
        return next(
            b
            for b in qcs.children()
            if hasattr(b, "metaObject")
            and b.metaObject().className() == "QtPrivate::QColorShowLabel"
        )

    def replace_qcolorshowlabel(self, widget: widgets.QWidget):
        # Find the dialog widget used to display the current
        # color, so we can replace it with our implementation
        qcs = self.get_qcolorshower()
        qcsl = self.get_qcolorshowlabel()
        qcs.layout().replaceWidget(qcsl, widget)
        # Make sure it doesn't receive signals while hidden
        qcsl.blockSignals(True)
        qcsl.hide()
        widget.show()

    def use_alpha_channel(self, value: bool = True):
        self.setOption(self.ColorDialogOption.ShowAlphaChannel, value)

    def hide_buttons(self, value: bool = True):
        self.setOption(self.ColorDialogOption.NoButtons, value)

    def use_native_dialog(self, value: bool = True):
        self.setOption(self.ColorDialogOption.DontUseNativeDialog, not value)

    @classmethod
    def get_custom_colors(cls) -> list[gui.Color]:
        return [gui.Color(cls.customColor(i)) for i in range(cls.customCount())]

    @classmethod
    def set_custom_colors(cls, colors_: list[datatypes.ColorType]):
        num = min(len(colors_), cls.customCount())
        for i in range(num):
            color = colors.get_color(colors_[i])
            cls.setCustomColor(i, color)


class CPAlphaShowLabel(widgets.Label):
    """Label which displays the currently-active color using checkerboard alpha.

    A replacement for QColorDialog's QColorShowLabel.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Length in pixels of a side of the checkerboard squares
        # (Pattern is made up of 2x2 squares, total size 2n x 2n)
        self.checkerboard_size = 8
        # Start out transparent by default
        self.color = self.parent().currentColor()
        self.parent().currentColorChanged.connect(self.update_color)
        self.pattern = gui.Pixmap.create_checkerboard_pattern(
            self.checkerboard_size, "#aaa", "#ccc"
        )

    def update_color(self, color: gui.QColor):
        self.color = color
        self.repaint()

    def paintEvent(self, event):
        """Show the current color using checkerboard alpha."""
        event.accept()
        with gui.Painter(self) as p:
            p.set_pen(None)
            if self.color.alphaF() < 1.0:
                p.drawTiledPixmap(event.rect(), self.pattern, QtCore.QPoint(4, 4))
            p.fillRect(event.rect(), self.color)


if __name__ == "__main__":
    app = widgets.app()
    dlg = ColorDialog()
    dlg.setOptions(ColorDialog.ColorDialogOption.ShowAlphaChannel)  # type: ignore
    label = CPAlphaShowLabel(dlg)
    dlg.replace_qcolorshowlabel(label)
    cs = dlg.get_qcolorshower()
    dlg.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import logging

import requests

from prettyqt import core


logger = logging.getLogger(__name__)


class WorkerSignals(core.Object):
    """TrayMenus' communication bus."""

    download_finished = core.Signal(bytes, str)  # response, url
    download_failed = core.Signal(str, str)  # msg, url


class DownloadWorker(core.QRunnable):
    def __init__(self, url: str, timeout: int = 30) -> None:
        super().__init__()
        self.url = url
        self.timeout = timeout
        self.signals = WorkerSignals()

    @core.Slot()
    def run(self) -> None:
        try:
            req = requests.get(self.url, verify=True, timeout=self.timeout)
        except Exception as e:
            msg = f"Exception '{e}' during download of '{self.url}'"
            logger.exception(msg)
            self.signals.download_failed.emit(msg, self.url)
        else:
            self.signals.download_finished.emit(req.content, self.url)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    pool = core.ThreadPool()
    worker = DownloadWorker("https://api.thecatapi.com/v1/images/search")
    worker.signals.download_finished.connect(print)
    worker.signals.download_finished.connect(app.quit)
    pool.start(worker)
    app.exec()



from __future__ import annotations

import logging

from mknodes.info import linkprovider


logger = logging.getLogger(__name__)


class QtLinkProvider(linkprovider.LinkProvider):
    def link_for_klass(self, kls: type):
        mod_path = kls.__module__
        if mod_path.startswith("prettyqt"):
            return linkprovider.linked(kls.__qualname__)
        return super().link_for_klass(kls)


if __name__ == "__main__":
    provider = QtLinkProvider()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, gui


if TYPE_CHECKING:
    from prettyqt.qt import QtWidgets


class Spin:
    def __init__(
        self, parent_widget: QtWidgets.QWidget, interval: int = 10, step: int = 1
    ):
        self.parent_widget = parent_widget
        self.interval = interval
        self.step = step
        self.info: dict[QtWidgets.QWidget, tuple[core.Timer, int, int]] = {}

    def _update(self):
        if self.parent_widget not in self.info:
            return
        timer, angle, step = self.info[self.parent_widget]

        if angle >= 360:  # noqa: PLR2004
            angle = 0

        angle += step
        self.info[self.parent_widget] = timer, angle, step
        self.parent_widget.update()

    def setup(self, painter: gui.Painter, rect: core.QRect):
        if self.parent_widget not in self.info:
            timer = core.Timer(self.parent_widget, timeout=self._update)
            self.info[self.parent_widget] = timer, 0, self.step
            timer.start(self.interval)
        else:
            timer, angle, self.step = self.info[self.parent_widget]
            with painter.offset_by(int(rect.width() * 0.5), int(rect.height() * 0.5)):
                painter.rotate(angle)


class Pulse(Spin):
    def __init__(self, parent_widget: QtWidgets.QWidget):
        super().__init__(parent_widget, interval=300, step=45)



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict, get_repr


SharedMemoryErrorStr = Literal[
    "none",
    "permission_denied",
    "invalid_size",
    "key_error",
    "already_exists",
    "not_found",
    "lock_error",
    "out_of_resources",
    "unknown",
]

SHARED_MEMORY_ERROR: bidict[
    SharedMemoryErrorStr, core.QSharedMemory.SharedMemoryError
] = bidict(
    none=core.QSharedMemory.SharedMemoryError.NoError,
    permission_denied=core.QSharedMemory.SharedMemoryError.PermissionDenied,
    invalid_size=core.QSharedMemory.SharedMemoryError.InvalidSize,
    key_error=core.QSharedMemory.SharedMemoryError.KeyError,
    already_exists=core.QSharedMemory.SharedMemoryError.AlreadyExists,
    not_found=core.QSharedMemory.SharedMemoryError.NotFound,
    lock_error=core.QSharedMemory.SharedMemoryError.LockError,
    out_of_resources=core.QSharedMemory.SharedMemoryError.OutOfResources,
    unknown=core.QSharedMemory.SharedMemoryError.UnknownError,
)


class SharedMemory(core.ObjectMixin, core.QSharedMemory):
    """Access to a shared memory segment."""

    def __repr__(self):
        return get_repr(self, self.key())

    def __reduce__(self):
        return type(self), (self.key())

    def get_error(self) -> SharedMemoryErrorStr:
        return SHARED_MEMORY_ERROR.inverse[self.error()]



from __future__ import annotations

from prettyqt import core


class SequentialAnimationGroup(core.AnimationGroupMixin, core.QSequentialAnimationGroup):
    """Sequential group of animations."""

    def reverse(self):
        """Reverse animation in-place by switching start and end values."""
        for anim in reversed(list(self)):
            old_start = anim.startValue()
            old_end = anim.endValue()
            anim.setStartValue(old_end)
            anim.setEndValue(old_start)

    def reversed(self) -> SequentialAnimationGroup:
        """Return a reversed copy of the animation."""
        new = core.MetaObject(self.metaObject()).copy(self)
        for anim in reversed(list(self)):
            animation = core.MetaObject(anim.get_metaobject()).copy(anim)
            animation = animation.reversed()
            new.addAnimation(animation)
        return new

    def append_reversed(self) -> core.SequentialAnimationGroup:
        """Return copy of animation with appended reverse animation."""
        revers = self.reversed()
        animation = core.SequentialAnimationGroup()
        animation.addAnimation(self)
        animation.addAnimation(revers)
        return animation



from __future__ import annotations

from typing import TypeVar

from prettyqt import core, widgets
from prettyqt.utils import listdelegators


T = TypeVar("T", bound=widgets.QWidget)


class ScrollArea(widgets.AbstractScrollAreaMixin, widgets.QScrollArea):
    """Scrolling view onto another widget."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        viewport = widgets.Widget(self, object_name=f"{type(self).__name__}_viewport")
        self.setViewport(viewport)

    def __add__(self, other: widgets.QWidget | widgets.QLayout | list):
        self.add(other)
        return self

    def get_visible_widgets(
        self,
        typ: type[T] = widgets.QWidget,
        partial_allowed: bool = True,
        margin: int = 10,
        recursive: bool = True,
    ) -> listdelegators.ListDelegator[T]:
        """Return all widgets which are visible in the viewport.

        Results can be filtered by type and whether widget is fully or partially visible.
        A positive margin increases the area to search for widgets, a negative margin
        decreases it.
        """
        widget = self.widget()
        viewport = self.viewport()
        rect = viewport.rect().adjusted(-margin, -margin, margin, margin)
        found = []
        for w in widget.find_children(typ, recursive=recursive):
            top_left = w.mapTo(viewport, w.rect().topLeft())
            bottom_right = w.mapTo(viewport, w.rect().bottomRight())
            mapped = core.Rect(top_left, bottom_right)
            if (partial_allowed and rect.intersects(mapped)) or (
                not partial_allowed and rect.contains(mapped)
            ):
                found.append(w)
        return listdelegators.ListDelegator(found)

    def get_children(self) -> listdelegators.ListDelegator[widgets.QWidget]:
        return self.widget().layout().get_children()

    def set_widget(self, widget: widgets.QWidget):
        self.setWidget(widget)

    def add_widget(self, *args, **kwargs):
        self.widget().layout().addWidget(*args, **kwargs)

    def add_layout(self, *args, **kwargs):
        self.widget().layout().addLayout(*args, **kwargs)

    def add(
        self,
        item: widgets.QWidget | widgets.QLayout | list,
        stretch: float | None = None,
    ):
        match item:
            case widgets.QWidget():
                self.add_widget(item)
                if stretch:
                    self.widget().layout().setStretchFactor(self.count() - 1, stretch)
            case widgets.QLayout():
                widget = widgets.Widget(self)
                widget.set_layout(item)
                self.add_widget(widget)
                if stretch:
                    self.setStretchFactor(self.count() - 1, stretch)
            case list():
                for i in item:
                    self.add(i, stretch)
        return item


if __name__ == "__main__":
    app = widgets.app()
    area = ScrollArea()
    widget = widgets.Widget()
    widget.set_layout("vertical")
    area.set_widget(widget)
    for i in range(50):
        w = widgets.RadioButton(str(i))
        area.setWidgetResizable(True)
        area.add_widget(w)
        area.ensureWidgetVisible(w)
    area.show()
    a = area.v_scrollbar.fx["value"].animate(0, 300, reverse=True, single_shot=False)
    app.sleep(2)
    a.stop()
    app.exec()



"""Tests for `prettyqt` package."""

# from prettyqt import svgwidgets


# def test_svgwidget():
#     widget = svgwidgets.SvgWidget()
#     widget.load_file("")


# def test_svggraphicsitem():
#     item = svgwidgets.GraphicsSvgItem()
#     assert item



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from prettyqt import core, gui
from prettyqt.qt import QtWidgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class FileIconProvider(gui.AbstractFileIconProviderMixin, QtWidgets.QFileIconProvider):
    """File icons for the QFileSystemModel class."""

    def get_icon(
        self,
        typ: gui.abstractfileiconprovider.IconTypeStr
        | core.QFileInfo
        | datatypes.PathType,
    ) -> gui.QIcon:
        if isinstance(typ, os.PathLike | core.QFileInfo):
            param = core.FileInfo(typ)
        else:
            param = gui.abstractfileiconprovider.ICON_TYPE[typ]
        return self.icon(param)

    def use_custom_directory_icons(self, state: bool = True):
        opt = self.Option(0) if state else self.Option.DontUseCustomDirectoryIcons
        self.setOptions(opt)

    def uses_custom_directory_icons(self) -> bool:
        return self.options() == self.Option(0)


if __name__ == "__main__":
    import pathlib

    from prettyqt import widgets

    app = widgets.app()
    path = pathlib.Path.home()
    provider = FileIconProvider()
    provider.get_icon(path)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from prettyqt import gui


if TYPE_CHECKING:
    from collections.abc import Iterator
    from re import Pattern


class BaseRuleSyntaxHighlighter(gui.SyntaxHighlighter):
    RULES: ClassVar[list] = []

    @classmethod
    def iter_rules(cls) -> Iterator[tuple[Pattern, int, gui.TextCharFormat]]:
        for rule in cls.RULES:
            if isinstance(rule.compiled, list):
                for i in rule.compiled:
                    yield (i, rule.nth, rule.fmt)
            else:
                yield (rule.compiled, rule.nth, rule.fmt)

    def highlightBlock(self, text: str):
        """Apply syntax highlighting to the given block of text."""
        # Do other syntax formatting
        for expression, nth, fmt in self.iter_rules():
            for match in expression.finditer(text):
                span = match.span(nth)
                self.setFormat(span[0], span[1] - span[0], fmt)



from __future__ import annotations

from prettyqt import constants, core, itemmodels


class SliceFilterProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model to filter an item view based on python slicing syntax.

    Since slicing operations are bijective, this model can filter without
    looping through rows or columns. Thus, this should perform much better than a
    SortFilterProxyModel with a column filter. (O(1) instead of O(n))

    ### Example

    To filter out every second row, and cut off the the first two columns:

    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[::2, 2:].filter()
    table.show()
    # or
    indexer = (slice(None, None, 2), slice(2, None))
    proxy = itemmodels.SliceFilterProxyModel(indexer=indexer)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "slice_filter"
    ICON = "mdi.table-filter"

    def rowCount(self, index: core.ModelIndex | None = None) -> int:
        rowcount = super().rowCount()
        # TODO: not sure if slice.stop = 0 is covered correctly?
        return min(rowcount, self.get_row_slice().stop or rowcount)

    def columnCount(self, index: core.ModelIndex | None = None) -> int:
        colcount = super().columnCount()
        return min(colcount, self.get_column_slice().stop or colcount)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        """Map header data to proxy by calculating position from slice values.

        source pos = slice start + proxy pos * slice step)
        """
        is_horizontal = orientation == constants.HORIZONTAL
        rng = self.get_column_range() if is_horizontal else self.get_row_range()
        pos = rng.start + section * rng.step
        return super().headerData(pos, orientation, role)

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        source = self.sourceModel()
        if row < 0 or column < 0 or source is None:
            return core.ModelIndex()
        source_parent = self.mapToSource(parent)
        col_range = self.get_column_range()
        row_range = self.get_row_range()
        col_pos = col_range.start + (col_range.step * column)
        row_pos = row_range.start + (row_range.step * row)
        source_index = source.index(row_pos, col_pos, source_parent)
        return self.mapFromSource(source_index)

    def mapToSource(self, proxy_idx: core.ModelIndex) -> core.ModelIndex:
        """Map index to source by calculating position from slice values.

        source pos = slice start + proxy pos * slice step)
        """
        source = self.sourceModel()
        if source is None or not proxy_idx.isValid():
            return core.ModelIndex()
        col_range = self.get_column_range()
        row_range = self.get_row_range()
        col_pos = col_range.start + (col_range.step * proxy_idx.column())
        row_pos = row_range.start + (row_range.step * proxy_idx.row())
        return source.index(row_pos, col_pos)

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        """Map index from source by calculating position based on slice values.

        proxy pos = source pos - slice start / slice step
        """
        if self.sourceModel() is None or not source_index.isValid():
            return core.ModelIndex()
        row_pos = self.position_in_row_slice(source_index.row())
        col_pos = self.position_in_column_slice(source_index.column())
        return self.createIndex(row_pos, col_pos, source_index.internalPointer())


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    table = debugging.example_table()
    model = SliceFilterProxyModel(parent=table)
    model.set_column_slice(slice(1, 4))
    model.setSourceModel(table.model())
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core, positioning
from prettyqt.qt import QtLocation
from prettyqt.utils import bidict


FeatureTypeStr = Literal[
    "none",
    "toll",
    "highway",
    "public_transit",
    "ferry",
    "tunnel",
    "dirt_road",
    "parks",
    "motor_pool_lane",
    "traffic",
]

FEATURE_TYPES: bidict[FeatureTypeStr, QtLocation.QGeoRouteRequest.FeatureType] = bidict(
    none=QtLocation.QGeoRouteRequest.FeatureType.NoFeature,
    toll=QtLocation.QGeoRouteRequest.FeatureType.TollFeature,
    highway=QtLocation.QGeoRouteRequest.FeatureType.HighwayFeature,
    public_transit=QtLocation.QGeoRouteRequest.FeatureType.PublicTransitFeature,
    ferry=QtLocation.QGeoRouteRequest.FeatureType.FerryFeature,
    tunnel=QtLocation.QGeoRouteRequest.FeatureType.TunnelFeature,
    dirt_road=QtLocation.QGeoRouteRequest.FeatureType.DirtRoadFeature,
    parks=QtLocation.QGeoRouteRequest.FeatureType.ParksFeature,
    motor_pool_lane=QtLocation.QGeoRouteRequest.FeatureType.MotorPoolLaneFeature,
    traffic=QtLocation.QGeoRouteRequest.FeatureType.TrafficFeature,
)

FeatureWeightStr = Literal["neutral", "prefer", "require", "avoid", "disallow"]

FEATURE_WEIGHTS: bidict[FeatureWeightStr, QtLocation.QGeoRouteRequest.FeatureWeight] = (
    bidict(
        neutral=QtLocation.QGeoRouteRequest.FeatureWeight.NeutralFeatureWeight,
        prefer=QtLocation.QGeoRouteRequest.FeatureWeight.PreferFeatureWeight,
        require=QtLocation.QGeoRouteRequest.FeatureWeight.RequireFeatureWeight,
        avoid=QtLocation.QGeoRouteRequest.FeatureWeight.AvoidFeatureWeight,
        disallow=QtLocation.QGeoRouteRequest.FeatureWeight.DisallowFeatureWeight,
    )
)

ManeuverDetailStr = Literal["none", "basic"]

MANEUVER_DETAIL: bidict[ManeuverDetailStr, QtLocation.QGeoRouteRequest.ManeuverDetail] = (
    bidict(
        none=QtLocation.QGeoRouteRequest.ManeuverDetail.NoManeuvers,
        basic=QtLocation.QGeoRouteRequest.ManeuverDetail.BasicManeuvers,
    )
)

RouteOptimizationStr = Literal["shortest", "fastest", "most_economic", "most_scenic"]

ROUTE_OPTIMIZATION: bidict[
    RouteOptimizationStr, QtLocation.QGeoRouteRequest.RouteOptimization
] = bidict(
    shortest=QtLocation.QGeoRouteRequest.RouteOptimization.ShortestRoute,
    fastest=QtLocation.QGeoRouteRequest.RouteOptimization.FastestRoute,
    most_economic=QtLocation.QGeoRouteRequest.RouteOptimization.MostEconomicRoute,
    most_scenic=QtLocation.QGeoRouteRequest.RouteOptimization.MostScenicRoute,
)

SegmentDetailStr = Literal["none", "basic"]

SEGMENT_DETAIL: bidict[SegmentDetailStr, QtLocation.QGeoRouteRequest.SegmentDetail] = (
    bidict(
        none=QtLocation.QGeoRouteRequest.SegmentDetail.NoSegmentData,
        basic=QtLocation.QGeoRouteRequest.SegmentDetail.BasicSegmentData,
    )
)

TravelModeStr = Literal["car", "pedestrian", "bicycle", "public_transit", "truck"]

TRAVEL_MODE: bidict[TravelModeStr, QtLocation.QGeoRouteRequest.TravelMode] = bidict(
    car=QtLocation.QGeoRouteRequest.TravelMode.CarTravel,
    pedestrian=QtLocation.QGeoRouteRequest.TravelMode.PedestrianTravel,
    bicycle=QtLocation.QGeoRouteRequest.TravelMode.BicycleTravel,
    public_transit=QtLocation.QGeoRouteRequest.TravelMode.PublicTransitTravel,
    truck=QtLocation.QGeoRouteRequest.TravelMode.TruckTravel,
)


class GeoRouteRequest(QtLocation.QGeoRouteRequest):
    """Represents the params and restrictions defining a routing information request."""

    def get_waypoints(self) -> list[positioning.GeoCoordinate]:
        return [positioning.GeoCoordinate(wp) for wp in self.waypoints()]

    def get_exclude_areas(self) -> list[positioning.GeoRectangle]:
        return [positioning.GeoRectangle(wp) for wp in self.excludeAreas()]

    def get_departure_time(self) -> core.DateTime:
        return core.DateTime(self.departureTime())

    def set_feature_weight(
        self,
        feature: FeatureTypeStr | QtLocation.QGeoRouteRequest.FeatureType,
        weight: FeatureWeightStr | QtLocation.QGeoRouteRequest.FeatureWeight,
    ):
        """Set the feature weight.

        Args:
            feature: Feature type
            weight: Feature weight
        """
        self.setFeatureWeight(
            FEATURE_TYPES.get_enum_value(feature), FEATURE_WEIGHTS.get_enum_value(weight)
        )

    def get_feature_weight(
        self, feature: FeatureTypeStr | QtLocation.QGeoRouteRequest.FeatureType
    ) -> FeatureWeightStr:
        """Return current feature weight.

        Returns:
            Feature weight
        """
        return FEATURE_WEIGHTS.inverse[
            self.featureWeight(FEATURE_TYPES.get_enum_value(feature))
        ]

    def set_route_optimization(
        self,
        optimization: RouteOptimizationStr
        | QtLocation.QGeoRouteRequest.RouteOptimization,
    ):
        """Set the route optimization.

        Args:
            optimization: Route optimization
        """
        self.setRouteOptimization(ROUTE_OPTIMIZATION.get_enum_value(optimization))

    def get_route_optimization(self) -> RouteOptimizationStr:
        """Return current route optimization.

        Returns:
            Route optimization
        """
        return ROUTE_OPTIMIZATION.inverse[self.routeOptimization()]

    def get_travel_modes(self) -> list[TravelModeStr]:
        return TRAVEL_MODE.get_list(self.travelModes())

    def set_travel_modes(self, *mode: TravelModeStr):
        flags = TRAVEL_MODE.merge_flags(mode)
        self.setTravelModes(flags)

    def get_feature_types(self) -> list[FeatureTypeStr]:
        return [k for k, v in FEATURE_TYPES.items() for t in self.featureTypes() if v & t]


if __name__ == "__main__":
    request = GeoRouteRequest()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import constants, widgets
from prettyqt.utils import bidict, datatypes


if TYPE_CHECKING:
    import datetime


SelectionModeStr = Literal["none", "single"]

SELECTION_MODE: bidict[SelectionModeStr, widgets.QCalendarWidget.SelectionMode] = bidict(
    none=widgets.QCalendarWidget.SelectionMode.NoSelection,
    single=widgets.QCalendarWidget.SelectionMode.SingleSelection,
)

VerticalHeaderFormatStr = Literal["none", "week_numbers"]

VERTICAL_HEADER_FORMAT: bidict[
    VerticalHeaderFormatStr, widgets.QCalendarWidget.VerticalHeaderFormat
] = bidict(
    none=widgets.QCalendarWidget.VerticalHeaderFormat.NoVerticalHeader,
    week_numbers=widgets.QCalendarWidget.VerticalHeaderFormat.ISOWeekNumbers,
)

HorizontalHeaderFormatStr = Literal["single_letter", "short", "long", "none"]

HORIZONTAL_HEADER_FORMAT: bidict[
    HorizontalHeaderFormatStr, widgets.QCalendarWidget.HorizontalHeaderFormat
] = bidict(
    single_letter=widgets.QCalendarWidget.HorizontalHeaderFormat.SingleLetterDayNames,
    short=widgets.QCalendarWidget.HorizontalHeaderFormat.ShortDayNames,
    long=widgets.QCalendarWidget.HorizontalHeaderFormat.LongDayNames,
    none=widgets.QCalendarWidget.HorizontalHeaderFormat.NoHorizontalHeader,
)


class CalendarWidget(widgets.WidgetMixin, widgets.QCalendarWidget):
    """Monthly based calendar widget allowing the user to select a date."""

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "firstDayOfWeek": constants.DAY_OF_WEEK,
            "horizontalHeaderFormat": HORIZONTAL_HEADER_FORMAT,
            "verticalHeaderFormat": VERTICAL_HEADER_FORMAT,
            "selectionMode": SELECTION_MODE,
        }
        return maps

    def get_date(self) -> datetime.date:
        return self.selectedDate().toPython()

    def get_value(self) -> datetime.date:
        return self.get_date()

    def set_value(self, value: datatypes.DateType):
        self.setSelectedDate(datatypes.to_date(value))

    def set_range(
        self,
        lower: datatypes.DateType,
        upper: datatypes.DateType,
    ):
        self.setMinimumDate(datatypes.to_date(lower))
        self.setMaximumDate(datatypes.to_date(upper))

    def set_selection_mode(
        self, mode: SelectionModeStr | widgets.QCalendarWidget.SelectionMode | None
    ):
        """Set selection mode for given calendar widget.

        Args:
            mode: selection mode to use
        """
        if mode is None:
            mode = "none"
        self.setSelectionMode(SELECTION_MODE.get_enum_value(mode))

    def get_selection_mode(self) -> SelectionModeStr:
        """Return current selection mode.

        Returns:
            selection mode
        """
        return SELECTION_MODE.inverse[self.selectionMode()]


if __name__ == "__main__":
    app = widgets.app()
    w = CalendarWidget()
    w.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.qt import QtOpenGLWidgets
from prettyqt.utils import bidict


TARGET_BUFFER = bidict(
    left_buffer=QtOpenGLWidgets.QOpenGLWidget.TargetBuffer.LeftBuffer,
    right_buffer=QtOpenGLWidgets.QOpenGLWidget.TargetBuffer.RightBuffer,
)

TargetBufferStr = Literal["left_buffer", "right_buffer"]


UPDATE_BEHAVIOR = bidict(
    no_partial=QtOpenGLWidgets.QOpenGLWidget.UpdateBehavior.NoPartialUpdate,
    partial=QtOpenGLWidgets.QOpenGLWidget.UpdateBehavior.PartialUpdate,
)

UpdateBehaviorStr = Literal["no_partial", "partial"]


class OpenGLWidget(widgets.WidgetMixin, QtOpenGLWidgets.QOpenGLWidget):
    """Widget for rendering OpenGL graphics."""

    def set_update_behavior(
        self, behavior: UpdateBehaviorStr | QtOpenGLWidgets.QOpenGLWidget.UpdateBehavior
    ):
        """Set update behavior.

        Args:
            behavior: update behavior to use
        """
        self.setUpdateBehavior(UPDATE_BEHAVIOR.get_enum_value(behavior))

    def get_update_behavior(self) -> UpdateBehaviorStr | None:
        """Return current update behavior.

        Returns:
            update behavior
        """
        return UPDATE_BEHAVIOR.inverse[self.updateBehavior()]

    def get_current_target_buffer(self) -> TargetBufferStr:
        """Return current target buffer.

        Returns:
            target buffer
        """
        return TARGET_BUFFER.inverse[self.currentTargetBuffer()]


if __name__ == "__main__":
    app = widgets.app()
    widget = OpenGLWidget()
    widget.show()
    app.exec()



"""Tests for `prettyqt` package."""

import pytest

from prettyqt import core, positioning
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


def test_geoaddress():
    address = positioning.GeoAddress()
    address.setCity("Test")
    assert str(address) == "Test"


def test_geolocation():
    location = positioning.GeoLocation()
    location.get_address()
    location.get_coordinate()
    location.get_bounding_shape()


def test_geocoordinate():
    coord = positioning.GeoCoordinate(11, 11)
    print(str(coord))
    assert repr(coord) == "GeoCoordinate(11.0, 11.0)"
    assert coord.get_type() == "2d"
    assert bool(coord)


def test_geocircle():
    coord = positioning.GeoCoordinate(1, 1)
    circle = positioning.GeoCircle(coord, 0.5)
    print(str(circle))
    assert circle.get_center() == coord
    assert repr(circle) == "GeoCircle(GeoCoordinate(1.0, 1.0), 0.5)"


def test_georectangle():
    coord1 = positioning.GeoCoordinate(1, 1)
    coord2 = positioning.GeoCoordinate(5, 5)
    rect = positioning.GeoRectangle(coord1, coord2)
    assert repr(rect) == "GeoRectangle(GeoCoordinate(1.0, 1.0), GeoCoordinate(5.0, 5.0))"


def test_geopath():
    coord1 = positioning.GeoCoordinate(1, 1)
    coord2 = positioning.GeoCoordinate(5, 5)
    path = positioning.GeoPath()
    path += coord1
    path += coord2
    path[1] = coord2
    assert path[1] == coord2
    assert repr(path) == "GeoPath(<2 points>)"
    assert positioning.GeoCoordinate(1, 1) in path
    del path[1]
    assert len(path) == 1


def test_geopolygon():
    coord1 = positioning.GeoCoordinate(1, 1)
    coord2 = positioning.GeoCoordinate(5, 5)
    path = positioning.GeoPolygon()
    path += coord1
    path += coord2
    path[1] = coord2
    assert path[1] == coord2
    assert repr(path) == "GeoPolygon(<2 points>)"
    del path[1]
    assert len(path) == 1
    str(path)


def test_geopositioninfo():
    info = positioning.GeoPositionInfo()
    info["direction"] = 10.0
    assert "direction" in info
    assert info["direction"] == 10.0
    del info["direction"]
    assert not bool(info)
    info.get_coordinate()
    info.get_timestamp()


def test_geosatelliteinfo():
    info = positioning.GeoSatelliteInfo()
    info["elevation"] = 1.0
    assert info["elevation"] == 1.0
    assert "elevation" in info
    info.setSatelliteIdentifier(1)
    assert int(info) == 1
    info.set_satellite_system("gps")
    with pytest.raises(InvalidParamError):
        info.set_satellite_system("test")
    assert info.get_satellite_system() == "gps"
    del info["elevation"]


def test_geoareamonitorsource():
    obj = core.Object()
    source = positioning.GeoAreaMonitorSource(obj)
    str(source)
    # assert source.get_error() == "none"


def test_geoareamonitorinfo():
    info = positioning.GeoAreaMonitorInfo()
    str(info)
    info.get_expiration()
    info.get_area()


def test_nmeapositioninfosource():
    obj = QtCore.QObject()
    source = positioning.NmeaPositionInfoSource("real_time", obj)
    print(repr(source))
    assert source.get_update_mode() == "real_time"
    assert source.get_error() == "unknown_source"
    source.set_preferred_positioning_methods("satellite")
    assert source.get_preferred_positioning_methods() == ["satellite", "all"]
    with pytest.raises(InvalidParamError):
        source.set_preferred_positioning_methods("test")
    source.get_supported_positioning_methods()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, core, widgets


if TYPE_CHECKING:
    from collections.abc import Iterable


class CycleWidget(widgets.ListWidget):
    current_item_changed = core.Signal(widgets.QListWidgetItem)

    def __init__(
        self,
        items: Iterable,
        item_size: core.QSize,
        align=constants.AlignmentFlag.AlignCenter,
        parent=None,
    ):
        super().__init__(parent=parent)
        self.item_size = item_size
        self.align = align

        self.up_button = widgets.ToolButton(
            parent=self, clicked=self.scroll_up, visible=False
        )
        self.up_button.set_icon("mdi.arrow-up-bold-outline")
        self.down_button = widgets.ToolButton(
            parent=self, clicked=self.scroll_down, visible=False
        )
        self.down_button.set_icon("mdi.arrow-down-bold-outline")
        self.origin_items = list(items)
        self.visible_number = 5
        self.setItems(items)
        self.set_vertical_scroll_mode("pixel")
        self.set_scrollbar_smooth(True, animation_duration=250)
        self.v_scrollbar.hide()
        self.set_viewport_margins(0)
        self.setFixedSize(item_size.width() + 8, item_size.height() * self.visible_number)

        self.set_scrollbar_policy("always_off")
        self.itemClicked.connect(self._on_item_clicked)

    def setItems(self, items: list[str]):
        self.clear()
        self._create_items(items)

    def _create_items(self, items: list[str]):
        N = len(items)  # noqa: N806
        self.is_cycle = self.visible_number < N

        if self.is_cycle:
            for _ in range(2):
                self._add_column_items(items)

            self._current_index = len(items)
            item = self.item(self.currentIndex() - self.visible_number // 2)
            super().scroll_to_item(item, "position_at_top")
        else:
            n = self.visible_number // 2  # add empty items to enable scrolling

            self._add_column_items([""] * n, True)
            self._add_column_items(items)
            self._add_column_items([""] * n, True)

            self._current_index = n

    def _add_column_items(self, items: list[str], disabled: bool = False):
        for i in items:
            item = widgets.ListWidgetItem(str(i), self)
            item.setSizeHint(self.item_size)
            item.setTextAlignment(self.align | constants.AlignmentFlag.AlignVCenter)
            if disabled:
                item.setFlags(constants.ItemFlag.NoItemFlags)

            self.addItem(item)

    def _on_item_clicked(self, item: widgets.ListWidgetItem):
        self.setCurrentIndex(self.row(item))
        self.scrollToItem(self.currentItem())

    def setSelectedItem(self, text: str):
        items = self.findItems(text, constants.MatchFlag.MatchExactly)
        if not items:
            return

        idx = self.row(items[1]) if len(items) >= 2 else self.row(items[0])  # noqa: PLR2004
        self.setCurrentIndex(idx)
        super().scroll_to_item(self.currentItem(), "position_at_center")

    def scrollToItem(
        self,
        item: widgets.QListWidgetItem,
        hint=widgets.ListWidget.ScrollHint.PositionAtCenter,
    ):
        index = self.row(item)
        y = item.sizeHint().height() * (index - self.visible_number // 2)
        self.v_scrollbar.scroll_to(y)
        self.clearSelection()
        item.setSelected(False)

        self.current_item_changed.emit(item)

    def wheelEvent(self, e):
        if e.angleDelta().y() < 0:
            self.scroll_down()
        else:
            self.scroll_up()

    def scroll_down(self):
        """Scroll down an item."""
        self.setCurrentIndex(self.currentIndex() + 1)
        self.scrollToItem(self.currentItem())

    def scroll_up(self):
        """Scroll up an item."""
        self.setCurrentIndex(self.currentIndex() - 1)
        self.scrollToItem(self.currentItem())

    def enterEvent(self, e):
        self.up_button.show()
        self.down_button.show()

    def leaveEvent(self, e):
        self.up_button.hide()
        self.down_button.hide()

    def resizeEvent(self, e):
        self.up_button.resize(self.width(), 34)
        self.down_button.resize(self.width(), 34)
        self.down_button.move(0, self.height() - 34)

    def keyPressEvent(self, e):
        match e.key():
            case constants.Key.Key_Down:
                self.scroll_down()
            case constants.Key.Key_Up:
                self.scroll_up()
            case _:
                super().keyPressEvent(e)

    def currentItem(self):
        return self.item(self.currentIndex())

    def currentIndex(self):
        return self._current_index

    def setCurrentIndex(self, index: int):
        if not self.is_cycle:
            n = self.visible_number // 2
            self._current_index = max(n, min(n + len(self.origin_items) - 1, index))
            return
        N = self.count() // 2  # noqa: N806
        m = (self.visible_number + 1) // 2
        self._current_index = index
        # scroll to center to achieve circular scrolling
        if index >= self.count() - m:
            self._current_index = N + index - self.count()
            item = self.item(self.currentIndex() - 1)
        elif index <= m - 1:
            self._current_index = N + index
            item = self.item(N + index + 1)
        else:
            return
        super().scrollToItem(item, self.ScrollHint.PositionAtCenter)


if __name__ == "__main__":
    app = widgets.app()
    widget = CycleWidget([str(i) for i in range(30)], core.QSize(50, 50))
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import widgets


class ColumnView(widgets.AbstractItemViewMixin, widgets.QColumnView):
    """Model/view implementation of a column view."""


if __name__ == "__main__":
    app = widgets.app()
    view = ColumnView()
    view.parent()
    view.show()
    app.exec()



from __future__ import annotations

from dataclasses import dataclass
import json
import os

from typing import Literal, overload, TYPE_CHECKING

# Third party imports
from prettyqt import gui, paths
from prettyqt.iconprovider.iconic_font import FontError, IconicFont, set_global_defaults
from prettyqt.qt import QtGui

if TYPE_CHECKING:
    import pathlib
    from prettyqt.utils import datatypes


key_type = tuple[str | None, str | None, bool]
icon_cache: dict[key_type, QtGui.QIcon] = {}
# Linux packagers, please set this to True if you want to make qtawesome
# use system fonts
SYSTEM_FONTS = False
PREFIXES = ("mdi.", "mdi6.", "fa5.", "fa5s.", "ei.", "ph.")


def hook(obj: dict) -> dict:
    result = {}
    for key, val in obj.items():
        try:
            result[key] = chr(int(val, 16))
        except ValueError as e:
            msg = f"Failed to load character {key}:{obj[key]}"
            raise FontError(msg) from e
    return result


@dataclass
class IconFont:
    """Class describing an Icon font."""

    prefix: str
    font_path: str
    charmap_path: str
    md5: str
    stylename: str | None = None
    path: pathlib.Path = paths.ICON_FONT_PATH

    def __post_init__(self):
        hash_val = None if SYSTEM_FONTS else self.md5
        id_ = gui.FontDatabase.add_font(self.path / self.font_path, ttf_hash=hash_val)
        loaded_font_families = gui.FontDatabase.applicationFontFamilies(id_)
        self.font_id = id_
        self.font_name = loaded_font_families[0]
        with (self.path / self.charmap_path).open("r") as codes:
            self.charmap = json.load(codes, object_hook=hook)

    def __dir__(self):
        icons = [i.replace("-", "_") for i in self.charmap]
        return list(super().__dir__()) + icons

    def __getattr__(self, name: str):
        return f"{self.prefix}.{name.replace('_', '-')}"

    def is_valid(self) -> bool:
        return len(gui.FontDatabase.applicationFontFamilies(self.font_id)) > 0

    def get_font(self, size: float) -> gui.Font:
        font = gui.Font(self.font_name)
        font.setPixelSize(round(size))
        if self.stylename:  # solid style
            font.setStyleName(self.stylename)
        return font


_resource: dict[str, IconicFont | None] = {"iconic": None}


def _instance() -> IconicFont:
    """Return the singleton instance of IconicFont.

    Functions ``icon``, ``load_font``, ``charmap``, ``font`` and
    ``set_defaults`` all rebind to methods of the singleton instance of IconicFont.
    """
    if _resource["iconic"] is not None and _resource["iconic"].has_valid_font_ids():
        return _resource["iconic"]
    font_awesome_5 = IconFont(
        prefix="fa5",
        font_path="fontawesome5-regular-webfont.ttf",
        charmap_path="fontawesome5-regular-webfont-charmap.json",
        md5="dc47e4089f5bcb25f241bdeb2de0fb58",
    )

    font_awesome_5_solid = IconFont(
        prefix="fa5s",
        font_path="fontawesome5-solid-webfont.ttf",
        charmap_path="fontawesome5-solid-webfont-charmap.json",
        md5="5de19800fc9ae73438c2e5c61d041b48",
        stylename="Solid",
    )

    font_awesome_5_brands = IconFont(
        prefix="fa5b",
        font_path="fontawesome5-brands-webfont.ttf",
        charmap_path="fontawesome5-brands-webfont-charmap.json",
        md5="513aa607d398efaccc559916c3431403",
    )

    elusive_icons = IconFont(
        prefix="ei",
        font_path="elusiveicons-webfont.ttf",
        charmap_path="elusiveicons-webfont-charmap.json",
        md5="207966b04c032d5b873fd595a211582e",
    )

    material_design_5_icons = IconFont(
        prefix="mdi",
        font_path="materialdesignicons5-webfont.ttf",
        charmap_path="materialdesignicons5-webfont-charmap.json",
        md5="b7d40e7ef80c1d4af6d94902af66e524",
    )

    material_design_6_icons = IconFont(
        prefix="mdi6",
        font_path="materialdesignicons6-webfont.ttf",
        charmap_path="materialdesignicons6-webfont-charmap.json",
        md5="9a2f455e7cbce011368aee95d292613b",
    )

    phosphor_icons = IconFont(
        prefix="ph",
        font_path="phosphor.ttf",
        charmap_path="phosphor-charmap.json",
        md5="5b8dc57388b2d86243566b996cc3a789",
    )

    remix_icons = IconFont(
        prefix="ri",
        font_path="remixicon.ttf",
        charmap_path="remixicon-charmap.json",
        md5="888e61f04316f10bddfff7bee10c6dd0",
    )

    codicon_icons = IconFont(
        prefix="msc",
        font_path="codicon.ttf",
        charmap_path="codicon-charmap.json",
        md5="d8565ee605ac4d2fa71fe018863337ca",
    )
    iconic = IconicFont(
        font_awesome_5,
        font_awesome_5_brands,
        font_awesome_5_solid,
        elusive_icons,
        material_design_5_icons,
        material_design_6_icons,
        phosphor_icons,
        remix_icons,
        codicon_icons,
    )
    _resource["iconic"] = iconic
    return iconic


def reset_cache():
    if _resource["iconic"] is not None:
        _resource["iconic"].icon_cache.clear()


def _icon(*names, **kwargs) -> QtGui.QIcon:
    """Return a QIcon object corresponding to the provided icon name(s).

    This function is the main interface of qtawesome.

    It can be used to create a QIcon instance from a single glyph
    or from a list of glyphs that are displayed on the top of each other.
    Such icon stacks are generally used to combine multiple glyphs to make
    more complex icons.

    Glyph names are specified by strings, of the form ``prefix.name``.
    The ``prefix`` corresponds to the font to be used and ``name`` is the
    name of the icon.

     - The prefix corresponding to Font-Awesome 4.x is 'fa'
     - The prefix corresponding to Font-Awesome 5.x (regular) is 'fa5'
     - The prefix corresponding to Font-Awesome 5.x (solid) is 'fa5s'
     - The prefix corresponding to Font-Awesome 5.x (brands) is 'fa5b'
     - The prefix corresponding to Elusive-Icons is 'ei'
     - The prefix corresponding to Material-Design-Icons is 'mdi'

    When requesting a single glyph, options (such as color or positional offsets)
    can be passed as keyword arguments::

        import qtawesome as qta

        music_icon = qta.icon(
            'fa5s.music',
            color='blue',
            color_active='orange')

    When requesting multiple glyphs, the `options` keyword argument contains
    the list of option dictionaries to be used for each glyph::

        camera_ban = qta.icon('fa5s.camera', 'fa5s.ban', options=[{
                'scale_factor': 0.5,
                'active': 'fa5s.balance-scale'
            }, {
                'color': 'red',
                'opacity': 0.7
            }])

    Qt's ``QIcon`` object has four modes

        - ``Normal``: The user is not interacting with the icon, but the
          functionality represented by the icon is available.
        - ``Disabled``: The functionality corresponding to the icon is not
          available.
        - ``Active``: The functionality corresponding to the icon is available.
          The user is interacting with the icon, for example, moving the mouse
          over it or clicking it.
        - ``Selected``: The item represented by the icon is selected.

    The glyph for the Normal mode is the one specified with the main positional
    argument.

     - ``color``: icon color in the ``Normal`` mode.

    The following four options will apply to the icon regardless of the mode.

     - ``offset``: tuple (x, y) corresponding to the horizontal and vertical
       offsets for the glyph, specified as a proportion of the icon size.
     - ``animation``: animation object for the icon.
     - ``scale_factor``: multiplicative scale factor to be used for the glyph.

    The following options apply to the different modes of the icon

     - ``active``: name of the glyph to be used when the icon is ``Active``.
     - ``color_active``: the corresponding icon color.

     - ``disabled``: name of the glyph to be used when the icon is ``Disabled``.
     - ``color_disabled``: the corresponding icon color.

     - ``selected``: name of the glyph to be used when the icon is ``Selected``.
     - ``color_selected``: the corresponding icon color.

    Default values for these options can be specified via the function
    ``set_defaults``. If unspecified, the default defaults are::

        {
            'opacity': 1.0,
            'scale_factor': 1.0
            'color': QColor(50, 50, 50),
            'color_disabled': QColor(150, 150, 150),
        }

    If no default value is provided for ``active``, ``disabled`` or ``selected``
    the glyph specified for the ``Normal`` mode will be used.

    """
    return _instance().icon(*names, **kwargs)


def for_color(color: str | QtGui.QColor | QtGui.QBrush) -> gui.Icon:
    match color:
        case str():
            color = gui.Color(color)
        case gui.QBrush():
            color = color.color()
        case gui.QColor():
            pass
        case _:
            raise TypeError(color)
    if color.isValid():
        bitmap = gui.Pixmap(16, 16)
        bitmap.fill(color)
        return gui.Icon(bitmap)
    return gui.Icon(_icon("mdi.card-outline"))


def set_defaults(**kwargs):
    """Set default options for icons.

    The valid keyword arguments are:

    'active', 'animation', 'color', 'color_active', 'color_disabled',
    'color_selected', 'disabled', 'offset', 'scale_factor', 'selected'.

    """
    return set_global_defaults(**kwargs)


@overload
def get_icon(
    icon: datatypes.IconType, color: str | None = None, as_qicon: Literal[False] = False
) -> gui.Icon: ...


@overload
def get_icon(
    icon: datatypes.IconType, color: str | None = None, as_qicon: Literal[True] = True
) -> QtGui.QIcon: ...


def get_icon(
    icon: datatypes.IconType, color: str | None = None, as_qicon: bool = False
) -> QtGui.QIcon:
    """Get icon with given color.

    Qtawesome already caches icons, but since we construct our own subclassed icon,
    we cache, too.
    """
    if isinstance(icon, QtGui.QIcon):
        return icon if as_qicon else gui.Icon(icon)
    if isinstance(icon, os.PathLike):
        icon = os.fspath(icon)
    if (icon, color, as_qicon) in icon_cache:
        return icon_cache[(icon, color, as_qicon)]
    if isinstance(icon, str) and icon.startswith(PREFIXES):
        new = _icon(icon, color=color) if color is not None else _icon(icon)
    else:
        new = QtGui.QIcon(icon)  # type: ignore
    icon = new if as_qicon else gui.Icon(new)
    icon_cache[(icon, color, as_qicon)] = icon
    return icon


if __name__ == "__main__":
    from prettyqt import custom_widgets, widgets

    app = widgets.app()
    icon = _icon("mdi.folder", vflip=True)
    # state = icon.__getstate__()
    # icon_2 = gui.Icon()
    # icon_2.__setstate__(state)
    widget = custom_widgets.IconWidget()
    widget.set_icon(icon)
    widget.set_icon_size(100)
    widget.show()
    app.exec()



"""Tests for `prettyqt` package."""

import dataclasses

# from importlib import metadata
import inspect
import re

import pytest

from prettyqt import core, itemmodels
from prettyqt.itemmodels import fsspecmodel
from prettyqt.utils import InvalidParamError


clsmembers = inspect.getmembers(itemmodels, inspect.isclass)
clsmembers = [tpl for tpl in clsmembers if core.QObject in tpl[1].mro()]


models = [tpl for tpl in clsmembers if not issubclass(tpl[1], core.QAbstractItemModel)]
proxies = [tpl for tpl in clsmembers if not issubclass(tpl[1], core.QAbstractProxyModel)]


def test_xmlmodel(qtmodeltester):
    xml = """<root>
<element key='value'>text</element>
<element><sub>text</sub></element>
<empty-element xmlns="http://testns/" />
</root>
"""
    model = itemmodels.XmlModel(xml)
    qtmodeltester.check(model, force_py=True)


def test_dataclassmodel(qtmodeltester):
    @dataclasses.dataclass
    class Test:
        a: int
        b: str

    test = Test(a=1, b="abc")
    model = itemmodels.DataClassModel([test])
    qtmodeltester.check(model, force_py=True)


def test_dataclassfieldsmodel(qtmodeltester):
    @dataclasses.dataclass
    class Test:
        a: int
        b: str

    test = Test(a=1, b="abc")
    model = itemmodels.DataClassFieldsModel(test)
    qtmodeltester.check(model, force_py=True)


def test_regexmatchesmodel(qtmodeltester):
    comp = re.compile("[0-9]")
    text = "aa356aa356"
    matches = list(comp.finditer(text))
    model = itemmodels.RegexMatchesModel(matches)
    qtmodeltester.check(model, force_py=True)


# def test_importlibtreemodel(qtmodeltester):
#     model = itemmodels.ImportlibTreeModel(metadata.distribution("prettyqt"))
#     qtmodeltester.check(model)
# itemmodels.ImportlibDistributionModel.from_package("prettyqt")
# itemmodels.ImportlibDistributionModel.from_system()


def test_jsonmodel(qtmodeltester):
    dct = {
        "lastName": "Smith",
        "age": 25,
        "address": {"streetAddress": "21 2nd Street", "postalCode": "10021"},
        "phoneNumber": [
            {"type": "home", "number": "212 555-1234"},
            {"type": "fax", "number": ("646 555-4567")},
        ],
    }
    model = itemmodels.JsonModel(dct)
    qtmodeltester.check(model, force_py=True)


def test_fsspecmodel(qtbot, qtmodeltester):
    model = fsspecmodel.FSSpecTreeModel()
    idx = model.index(0, 0)
    model.get_paths([idx])
    model.data(idx, model.Roles.FilePathRole)
    model.watch_for_changes(False)
    model.use_custom_icons(False)
    model.resolve_sym_links(False)
    model.set_name_filters(["test"], hide=True)
    model.set_filter("drives")
    with pytest.raises(InvalidParamError):
        model.set_filter("test")
    # qtmodeltester.check(model, force_py=True)



"""Tests for `prettyqt` package."""

import pathlib

from prettyqt import quickwidgets


def test_quickview():
    view = quickwidgets.QuickWidget()
    path = pathlib.Path.cwd() / "tests" / "qmltest.qml"
    view.set_source(path)
    assert view.get_source() == path
    assert view.get_status() == "ready"



"""syntaxhighlighters module.

contains some custom syntax highlighers
"""

from .highlightrule import HighlightRule
from .baserulesyntaxhighlighter import BaseRuleSyntaxHighlighter
from .jsonhighlighter import JsonHighlighter
from .markdownhighlighter import MarkdownHighlighter
from .pygmentshighlighter import PygmentsHighlighter
from .pythonhighlighter import PythonHighlighter
from .regexmatchhighlighter import RegexMatchHighlighter
from .selectedwordhighlighter import SelectedWordHighlighter
from .yamlhighlighter import YamlHighlighter


__all__ = [
    "BaseRuleSyntaxHighlighter",
    "HighlightRule",
    "JsonHighlighter",
    "MarkdownHighlighter",
    "PygmentsHighlighter",
    "PythonHighlighter",
    "RegexMatchHighlighter",
    "SelectedWordHighlighter",
    "YamlHighlighter",
]



# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "typed_ast",
# ]
# ///

import pathlib
import sys

from typed_ast import ast3


def add_parents(tree):
    for node in ast3.walk(tree):
        for child in ast3.iter_child_nodes(node):
            child.parent = node  # type: ignore


def find_enums(tree):
    for node in ast3.walk(tree):
        if not isinstance(node, ast3.Assign):
            continue
        if node.type_comment is None:
            continue
        if "." not in node.type_comment:
            continue
        if not node.type_comment.startswith("'"):
            continue
        comment = node.type_comment.strip("'")
        mod, cls = comment.rsplit(".", maxsplit=1)
        assert len(node.targets) == 1
        name = node.targets[0].id  # type: ignore
        yield (mod, cls, name)


def main():
    for filename in sys.argv[1:]:
        tree = ast3.parse(pathlib.Path(filename).read_text())
        for mod, cls, name in find_enums(tree):
            old = f"{mod}.{name}"
            new = f"{mod}.{cls}.{name}"
            print(f"{old} {new}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import contextlib

from prettyqt.qt import QtGui


class BackingStore(QtGui.QBackingStore):
    """Drawing area for QWindow."""

    @contextlib.contextmanager
    def paint_on_region(self, region: QtGui.QRegion):
        self.beginPaint(region)
        yield self
        self.endPaint()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    w = widgets.Widget()
    backingstore = BackingStore(w.backingStore())



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class ButtonDelegate(widgets.StyledItemDelegate):
    """Delegate to show a button inside a table cell."""

    ID = "button"

    def __init__(
        self,
        parent: widgets.QAbstractItemView,
        role: constants.ItemDataRole = constants.USER_ROLE,
    ):
        super().__init__(parent)
        parent.setMouseTracking(True)
        self.btn = widgets.PushButton(parent=parent)
        self.method_role = role
        self.btn.hide()
        self.current_edited_index = core.ModelIndex()
        parent.entered.connect(self.cell_entered)

    #     parent.viewport().installEventFilter(self)

    # def eventFilter(self, source, event) -> bool:
    #     if event.type() == event.Type.MouseMove:
    #         return True
    #     return super().eventFilter(source, event)

    def updateEditorGeometry(
        self,
        editor: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        editor.setGeometry(option.rect)

    def createEditor(
        self,
        parent: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        btn_callback = index.data(self.method_role)
        if btn_callback is None:
            return None
        return widgets.PushButton(parent=parent, text=index.data(), clicked=btn_callback)

    def setEditorData(self, editor: widgets.QWidget, index: core.ModelIndex):
        pass
        # editor.setProperty("test", "aa")
        # editor.setText(index.data())

    def setModelData(
        self,
        editor: widgets.QWidget,
        model: core.QAbstractItemModel,
        index: core.ModelIndex,
    ):
        pass
        # model.setData(index, editor.property("test"))

    def cell_entered(self, index: core.ModelIndex):
        # index = self.parent().indexFromItem(item)
        parent: widgets.QAbstractItemView = self.parent()  # type: ignore
        if parent.isPersistentEditorOpen(index):
            parent.closePersistentEditor(self.current_edited_index)
        if parent.itemDelegateForIndex(index) is self:
            # if index.data(self.method_role) is not None:
            parent.openPersistentEditor(index)
            parent.setCurrentIndex(index)
        self.current_edited_index = index

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        if index.data(self.method_role) is None:
            super().paint(painter, option, index)
            return
        self.btn.setGeometry(option.rect)
        self.btn.setText(index.data())
        if option.state & widgets.Style.StateFlag.State_MouseOver:
            painter.fillRect(option.rect, option.palette.highlight())
        if option.state & widgets.Style.StateFlag.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight())
        pixmap = self.btn.grab()
        painter.drawPixmap(option.rect, pixmap)
        # return super().paint(painter, option, index)

    # mouseOver = option.state in [73985, 73729, 8192]
    # if option.state & widgets.Style.StateFlag.State_MouseOver:
    #     painter.fillRect(option.rect, gui.Brush(gui.Color("yellow")))

    # def editorEvent(self, event, model, option, index):
    #     if event.type() == core.QEvent.Type.MouseButtonRelease:
    #         logger.info("Clicked on Item", index.row())
    #         return True
    #     elif event.type() == core.QEvent.Type.MouseButtonDblClick:
    #         logger.info("Double-Clicked on Item", index.row())
    #         return True
    #     else:
    #         return super().editorEvent(event, model, option, index)

    # @core.Slot()
    # def currentIndexChanged(self):
    #     self.commitData.emit(self.sender())


if __name__ == "__main__":
    app = widgets.app()
    table_widget = widgets.TableWidget(15, 4)
    # table_widget.set_delegate(StarDelegate(), column=1)
    table_widget.setEditTriggers(
        table_widget.EditTrigger.DoubleClicked  # type: ignore
        | table_widget.EditTrigger.SelectedClicked
    )
    table_widget.set_selection_behavior("rows")
    # table_widget.set_selection_mode(None)
    table_widget.setHorizontalHeaderLabels(["Title", "Rating"])
    for i in range(10):
        item_1 = widgets.TableWidgetItem(str(i))
        item_2 = widgets.TableWidgetItem()
        item_2.setData(constants.DISPLAY_ROLE, str(i))
        item_2.setData(constants.USER_ROLE, lambda: print("test"))
        table_widget[i, 0] = item_1
        table_widget[i, 1] = item_2

    table_widget.set_delegate("button", column=1)

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()

    app.exec()



from __future__ import annotations

from prettyqt.qt import QtCore


class RandomGenerator(QtCore.QRandomGenerator):
    """Allows one to obtain random values from a high-quality Random Number Generator."""



from __future__ import annotations

from prettyqt.qt import QtGui
from prettyqt.utils import get_repr


class TextBlockUserData(QtGui.QTextBlockUserData):
    """Storage for the user data associated with each line."""

    def __init__(self, **kwds):
        for key, value in kwds.items():
            setattr(self, key, value)
        super().__init__()

    def __repr__(self):
        kwargs = {i: getattr(self, i) for i in dir(self) if not i.startswith("__")}
        return get_repr(self, **kwargs)



from __future__ import annotations

from prettyqt import core, widgets


class GraphicsObjectMixin(core.ObjectMixin, widgets.GraphicsItemMixin):
    pass


class GraphicsObject(GraphicsObjectMixin, widgets.QGraphicsObject):
    """Base class for all graphics items that require signals, slots and properties."""



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator


mod = widgets.QWizard

WizardButtonStr = Literal[
    "back",
    "next",
    "commit",
    "finish",
    "cancel",
    "help",
    "custom_1",
    "custom_2",
    "custom_3",
]

WIZARD_BUTTON: bidict[WizardButtonStr, mod.WizardButton] = bidict(
    back=mod.WizardButton.BackButton,
    next=mod.WizardButton.NextButton,
    commit=mod.WizardButton.CommitButton,
    finish=mod.WizardButton.FinishButton,
    cancel=mod.WizardButton.CancelButton,
    help=mod.WizardButton.HelpButton,
    custom_1=mod.WizardButton.CustomButton1,
    custom_2=mod.WizardButton.CustomButton2,
    custom_3=mod.WizardButton.CustomButton3,
)

WizardOptionStr = Literal[
    "independent_pages",
    "ignore_subtitles",
    "extended_watermark_bitmap",
    "no_default_button",
    "no_back_button_on_start_page",
    "no_back_button_on_last_page",
    "disabled_back_button_on_last_page",
    "next_button_on_last_page",
    "finish_button_on_early_pages",
    "no_cancel_button",
    "cancel_button_on_left",
    "help_button",
    "help_button_on_right",
    "custom_button_1",
    "custom_button_2",
    "custom_button_3",
    "no_cancel_button_on_last_page",
]

WIZARD_OPTIONS: bidict[WizardOptionStr, mod.WizardOption] = bidict(
    independent_pages=mod.WizardOption.IndependentPages,
    ignore_subtitles=mod.WizardOption.IgnoreSubTitles,
    extended_watermark_bitmap=mod.WizardOption.ExtendedWatermarkPixmap,
    no_default_button=mod.WizardOption.NoDefaultButton,
    no_back_button_on_start_page=mod.WizardOption.NoBackButtonOnStartPage,
    no_back_button_on_last_page=mod.WizardOption.NoBackButtonOnLastPage,
    disabled_back_button_on_last_page=mod.WizardOption.DisabledBackButtonOnLastPage,
    next_button_on_last_page=mod.WizardOption.HaveNextButtonOnLastPage,
    finish_button_on_early_pages=mod.WizardOption.HaveFinishButtonOnEarlyPages,
    no_cancel_button=mod.WizardOption.NoCancelButton,
    cancel_button_on_left=mod.WizardOption.CancelButtonOnLeft,
    help_button=mod.WizardOption.HaveHelpButton,
    help_button_on_right=mod.WizardOption.HelpButtonOnRight,
    custom_button_1=mod.WizardOption.HaveCustomButton1,
    custom_button_2=mod.WizardOption.HaveCustomButton2,
    custom_button_3=mod.WizardOption.HaveCustomButton3,
    no_cancel_button_on_last_page=mod.WizardOption.NoCancelButtonOnLastPage,
)

WizardPixmapStr = Literal["watermark", "logo", "banner", "background"]

WIZARD_PIXMAP: bidict[WizardPixmapStr, mod.WizardPixmap] = bidict(
    watermark=mod.WizardPixmap.WatermarkPixmap,
    logo=mod.WizardPixmap.LogoPixmap,
    banner=mod.WizardPixmap.BannerPixmap,
    background=mod.WizardPixmap.BackgroundPixmap,
)

WizardStyleStr = Literal["classic", "modern", "mac", "aero"]

WIZARD_STYLE: bidict[WizardStyleStr, mod.WizardStyle] = bidict(
    classic=mod.WizardStyle.ClassicStyle,
    modern=mod.WizardStyle.ModernStyle,
    mac=mod.WizardStyle.MacStyle,
    aero=mod.WizardStyle.AeroStyle,
)


class WizardMixin(widgets.DialogMixin):
    custom_button_1_clicked = core.Signal()
    custom_button_2_clicked = core.Signal()
    custom_button_3_clicked = core.Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.customButtonClicked.connect(self._on_custom_button_clicked)

    def __getitem__(self, key: int) -> widgets.QWizardPage:
        p = self.page(key)
        if p is None:
            raise KeyError(key)
        return p

    def __setitem__(self, key: int, value: widgets.QWizardPage):
        return self.setPage(key, value)

    def __delitem__(self, key: int):
        if key not in self.pageIds():
            raise KeyError(key)
        return self.removePage(key)

    def __iter__(self) -> Iterator[widgets.QWizardPage]:
        return iter(self.page(i) for i in self.pageIds())

    def __add__(self, other: widgets.QWizardPage) -> Wizard:
        self.addPage(other)
        return self

    def _on_custom_button_clicked(self, which: int):
        match which:
            case 6:
                self.custom_button_1_clicked.emit()
            case 7:
                self.custom_button_2_clicked.emit()
            case 8:
                self.custom_button_3_clicked.emit()

    def add_widget_as_page(self, widget: widgets.QWidget) -> None:
        page = widgets.WizardPage(self)
        layout = page.set_layout("vertical")
        layout += widget

    def set_title_format(self, fmt: constants.TextFormatStr | constants.TextFormat):
        """Set the title format.

        Allowed values are "rich", "plain", "auto", "markdown"

        Args:
            fmt: title format to use
        """
        self.setTitleFormat(constants.TEXT_FORMAT.get_enum_value(fmt))

    def get_title_format(self) -> constants.TextFormatStr:
        """Return current title format.

        Possible values: "rich", "plain", "auto", "markdown"

        Returns:
            title format
        """
        return constants.TEXT_FORMAT.inverse[self.titleFormat()]

    def set_subtitle_format(self, fmt: constants.TextFormatStr | constants.TextFormat):
        """Set the subtitle format.

        Allowed values are "rich", "plain", "auto", "markdown"

        Args:
            fmt: subtitle format to use
        """
        self.setSubTitleFormat(constants.TEXT_FORMAT.get_enum_value(fmt))

    def get_subtitle_format(self) -> constants.TextFormatStr:
        """Return current subtitle format.

        Possible values: "rich", "plain", "auto", "markdown"

        Returns:
            subtitle format
        """
        return constants.TEXT_FORMAT.inverse[self.subTitleFormat()]

    def get_button(
        self, button_type: WizardButtonStr | mod.WizardButton
    ) -> widgets.QAbstractButton:
        return self.button(WIZARD_BUTTON.get_enum_value(button_type))

    def set_button_text(
        self, button_type: WizardButtonStr | mod.WizardButton, value: str
    ):
        """Set text for given button type.

        Args:
            button_type: button to get text from
            value: text to set

        """
        self.setButtonText(WIZARD_BUTTON.get_enum_value(button_type), value)

    def get_button_text(self, button_type: WizardButtonStr | mod.WizardButton) -> str:
        """Return text for given button type.

        Args:
            button_type: button to get text from

        Returns:
            Button text
        """
        return self.buttonText(WIZARD_BUTTON.get_enum_value(button_type))

    def set_pixmap(
        self, typ: WizardPixmapStr | mod.WizardPixmap, pixmap: gui.QPixmap | None
    ):
        if pixmap is None:
            pixmap = gui.QPixmap()
        self.setPixmap(WIZARD_PIXMAP.get_enum_value(typ), pixmap)

    def get_pixmap(self, typ: WizardPixmapStr | mod.WizardPixmap) -> gui.Pixmap | None:
        pix = gui.Pixmap(self.pixmap(WIZARD_PIXMAP.get_enum_value(typ)))
        return None if pix.isNull() else pix

    def set_wizard_style(self, style: WizardStyleStr | mod.WizardStyle):
        """Set the wizard style.

        Args:
            style: wizard style
        """
        self.setWizardStyle(WIZARD_STYLE.get_enum_value(style))

    def get_wizard_style(self) -> WizardStyleStr:
        """Return current wizard style.

        Returns:
            Wizard style
        """
        return WIZARD_STYLE.inverse[self.wizardStyle()]

    def set_option(self, option: WizardOptionStr | mod.WizardOption, value: bool):
        """Set option to given value.

        Args:
            option: option to use
            value: value to set
        """
        self.setOption(WIZARD_OPTIONS.get_enum_value(option), value)

    def get_option(self, option: WizardOptionStr | mod.WizardOption) -> bool:
        """Return the value assigned to option.

        Args:
            option: option to get

        Returns:
            option
        """
        return self.testOption(WIZARD_OPTIONS.get_enum_value(option))

    def set_custom_button(
        self, button: Literal[1, 2, 3], text: str | None, callback: Callable | None = None
    ):
        self.set_option(f"custom_button_{button}", text is not None)
        if text is None:
            return
        self.setButtonText(WIZARD_BUTTON[f"custom_{button}"], text)
        if callback:
            match button:
                case 1:
                    self.custom_button_1_clicked.connect(callback)
                case 2:
                    self.custom_button_2_clicked.connect(callback)
                case 3:
                    self.custom_button_3_clicked.connect(callback)


class Wizard(WizardMixin, widgets.QWizard):
    """Framework for wizards."""


if __name__ == "__main__":
    app = widgets.app()
    dlg = Wizard()
    dlg.add_widget_as_page(widgets.RadioButton("test"))
    dlg.set_custom_button(1, "1", lambda: print("1"))
    dlg.set_custom_button(2, "2", lambda: print("2"))
    dlg.set_custom_button(3, "3", lambda: print("3"))
    dlg.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict


mod = QtWebEngineCore.QWebEngineContextMenuRequest

EDIT_FLAGS = bidict(
    undo=mod.EditFlag.CanUndo,
    redo=mod.EditFlag.CanRedo,
    cut=mod.EditFlag.CanCut,
    copy=mod.EditFlag.CanCopy,
    paste=mod.EditFlag.CanPaste,
    delete=mod.EditFlag.CanDelete,
    select_all=mod.EditFlag.CanSelectAll,
    translate=mod.EditFlag.CanTranslate,
    edit_richly=mod.EditFlag.CanEditRichly,
)

EditFlagStr = Literal[
    "undo",
    "redo",
    "cut",
    "copy",
    "paste",
    "delete",
    "select_all",
    "translate",
    "edit_richly",
]

MEDIA_FLAGS = bidict(
    in_error=mod.MediaFlag.MediaInError,
    paused=mod.MediaFlag.MediaPaused,
    muted=mod.MediaFlag.MediaMuted,
    loop=mod.MediaFlag.MediaLoop,
    can_save=mod.MediaFlag.MediaCanSave,
    has_audio=mod.MediaFlag.MediaHasAudio,
    can_toggle_controls=mod.MediaFlag.MediaCanToggleControls,
    controls=mod.MediaFlag.MediaControls,
    can_print=mod.MediaFlag.MediaCanPrint,
    can_rotate=mod.MediaFlag.MediaCanRotate,
)

MediaFlagStr = Literal[
    "in_error",
    "paused",
    "muted",
    "loop",
    "can_save",
    "has_audio",
    "can_toggle_controls",
    "controls",
    "can_print",
    "can_rotate",
]

MEDIA_TYPES = bidict(
    none=mod.MediaType.MediaTypeNone,
    image=mod.MediaType.MediaTypeImage,
    video=mod.MediaType.MediaTypeVideo,
    audio=mod.MediaType.MediaTypeAudio,
    canvas=mod.MediaType.MediaTypeCanvas,
    file=mod.MediaType.MediaTypeFile,
    plugin=mod.MediaType.MediaTypePlugin,
)

MediaTypeStr = Literal["none", "image", "video", "audio", "canvas", "file", "plugin"]


class WebEngineContextMenuRequest(QtWebEngineCore.QWebEngineContextMenuRequest):
    def get_media_url(self) -> core.Url:
        return core.Url(self.mediaUrl())

    def get_link_url(self) -> core.Url:
        return core.Url(self.linkUrl())

    def get_media_type(self) -> MediaTypeStr:
        """Get media type."""
        return MEDIA_TYPES.inverse[self.mediaType()]

    def get_media_flags(self) -> list[MediaFlagStr]:
        """Get media flags."""
        return MEDIA_FLAGS.get_list(self.mediaFlags())

    def get_edit_flags(self) -> list[MediaFlagStr]:
        """Get edit flags."""
        return EDIT_FLAGS.get_list(self.editFlags())

    def can_undo(self) -> bool:
        """Returns whether undo is possible."""
        return bool(mod.EditFlag.CanUndo & self.editFlags())

    def can_redo(self) -> bool:
        """Returns whether redo is possible."""
        return bool(mod.EditFlag.CanRedo & self.editFlags())

    def can_cut(self) -> bool:
        """Returns whether cut is possible."""
        return bool(mod.EditFlag.CanCut & self.editFlags())

    def can_copy(self) -> bool:
        """Returns whether copy is possible."""
        return bool(mod.EditFlag.CanCopy & self.editFlags())

    def can_paste(self) -> bool:
        """Returns whether paste is possible."""
        return bool(mod.EditFlag.CanPaste & self.editFlags())

    def can_delete(self) -> bool:
        """Returns whether delete is possible."""
        return bool(mod.EditFlag.CanDelete & self.editFlags())

    def can_select_all(self) -> bool:
        """Returns whether select_all is possible."""
        return bool(mod.EditFlag.CanSelectAll & self.editFlags())

    def can_translate(self) -> bool:
        """Returns whether translate is possible."""
        return bool(mod.EditFlag.CanTranslate & self.editFlags())

    def can_edit_richly(self) -> bool:
        """Returns whether edit_richly is possible."""
        return bool(mod.EditFlag.CanEditRichly & self.editFlags())


if __name__ == "__main__":
    from prettyqt import webenginecore, widgets

    app = widgets.app()
    page = webenginecore.WebEnginePage()
    context_menu_data = page.get_context_menu_data()
    print(context_menu_data.can_copy())
    app.exec()



from __future__ import annotations

from collections.abc import Sequence
import logging
from typing import Any

from prettyqt import constants, core, itemmodels
from prettyqt.utils import bidict


logger = logging.getLogger(__name__)


class ModelIndexModel(itemmodels.ListMixin, core.AbstractTableModel):
    """Table model for displaying a the data of a list of ModelIndexes.

    ### Example:
    ```py
    # get indexes from some random model
    my_model = ParentClassTreeModel(widgets.QWidget, show_mro=True)
    indexes = list(my_model.iter_tree(fetch_more=True))
    model = ModelIndexModel(indexes=indexes)
    ```
    """

    SUPPORTS = Sequence[core.QModelIndex]
    ID = "modelindex"
    FIXED_HEADER = ["Path", "Row", "Column", "Flags"]

    def __init__(
        self,
        indexes: list[core.ModelIndex],
        **kwargs,
    ):
        self._use_model_roles = False
        super().__init__(**kwargs)
        self.items = indexes
        self._update_columns()

    def _update_columns(self):
        if self.items and self._use_model_roles:
            # we assume here that all indexes have some model.
            model = self.items[0].model()
            roles = bidict({i: v.data().decode() for i, v in model.roleNames().items()})
            self.role_mapping = roles.inverse
            self.role_headers = list(roles.values())
        else:
            self.role_mapping = {
                f"{k.capitalize()} role": v for k, v in constants.ITEM_DATA_ROLE.items()
            }
            self.role_headers = list(self.role_mapping.keys())

    def setup_delegates(self, view):
        for i, v in enumerate(self.role_mapping.values(), start=len(self.FIXED_HEADER)):
            view.set_delegate("editor", column=i, data_role=v)

    @property
    def headers(self):
        return self.FIXED_HEADER + self.role_headers

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (core.ModelIndex(), *_):
                return True
            case _:
                return False

    def columnCount(self, parent=None) -> int:
        return len(self.headers)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.headers[section]
            case constants.VERTICAL, constants.DISPLAY_ROLE:
                return str(section)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        idx = self.items[index.row()]
        header = self.headers[index.column()]
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                #  pieces = self.get_breadcrumbs_path()
                pieces = self.get_index_key(idx, include_column=True)
                return " / ".join(str(i) for i in pieces)
            case constants.DISPLAY_ROLE, 1:
                return idx.row()
            case constants.DISPLAY_ROLE, 2:
                return idx.column()
            case constants.DISPLAY_ROLE, 3:
                roles = constants.ITEM_FLAG.get_list(idx.flags())
                return " / ".join(roles)
            case constants.DISPLAY_ROLE, _:
                role = self.role_mapping[header]
                return repr(idx.data(role))
            case _ as role, _:
                # print(self.headers[column], role)
                return idx.data(role)

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        idx = self.items[index.row()]
        idx.model().setData(idx, value, role)
        self.update_row(index.row())
        return True

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        return super().flags(index) | constants.IS_EDITABLE
        # idx = self.items[index.row()]
        # return idx.flags()

    def set_use_model_roles(self, value: bool):
        self._use_model_roles = value

    def is_using_model_roles(self) -> bool:
        return self._use_model_roles

    use_model_roles = core.Property(
        bool,
        is_using_model_roles,
        set_use_model_roles,
        doc="Whether to use model.roleNames() for the columns.",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    with app.debug_mode():
        view = widgets.TableView()
        my_model = itemmodels.ParentClassTreeModel(widgets.QWidget, show_mro=True)
        indexes = list(my_model.iter_tree(fetch_more=True))
        model = ModelIndexModel(indexes=indexes, parent=view)
        for i, v in enumerate(model.role_mapping.values(), start=len(model.FIXED_HEADER)):
            view.set_delegate("editor", column=i, role=v)
        # view.set_delegate("editor")
        view.set_model(model)
        view.set_selection_behavior("rows")
        view.set_edit_triggers("all")
        view.show()
        view.resize(1000, 1000)
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import printsupport
from prettyqt.utils import bidict


PrintEnginePropertyKey = printsupport.QPrintEngine.PrintEnginePropertyKey

PRINT_ENGINE_PROPERTY_KEY = bidict(
    collate_copies=PrintEnginePropertyKey.PPK_CollateCopies,
    color_mode=PrintEnginePropertyKey.PPK_ColorMode,
    creator=PrintEnginePropertyKey.PPK_Creator,
    duplex=PrintEnginePropertyKey.PPK_Duplex,
    document_name=PrintEnginePropertyKey.PPK_DocumentName,
    font_embedding=PrintEnginePropertyKey.PPK_FontEmbedding,
    full_page=PrintEnginePropertyKey.PPK_FullPage,
    number_of_copies=PrintEnginePropertyKey.PPK_NumberOfCopies,
    orientation=PrintEnginePropertyKey.PPK_Orientation,
    output_file_name=PrintEnginePropertyKey.PPK_OutputFileName,
    page_order=PrintEnginePropertyKey.PPK_PageOrder,
    page_rect=PrintEnginePropertyKey.PPK_PageRect,
    page_size=PrintEnginePropertyKey.PPK_PageSize,
    paper_rect=PrintEnginePropertyKey.PPK_PaperRect,
    paper_source=PrintEnginePropertyKey.PPK_PaperSource,
    paper_sources=PrintEnginePropertyKey.PPK_PaperSources,
    paper_name=PrintEnginePropertyKey.PPK_PaperName,
    # paper_size=PrintEnginePropertyKey.PPK_PaperSize,
    printer_name=PrintEnginePropertyKey.PPK_PrinterName,
    printer_program=PrintEnginePropertyKey.PPK_PrinterProgram,
    resolution=PrintEnginePropertyKey.PPK_Resolution,
    selection_option=PrintEnginePropertyKey.PPK_SelectionOption,
    supported_resolutions=PrintEnginePropertyKey.PPK_SupportedResolutions,
    windows_page_size=PrintEnginePropertyKey.PPK_WindowsPageSize,
    custom_paper_size=PrintEnginePropertyKey.PPK_CustomPaperSize,
    page_margins=PrintEnginePropertyKey.PPK_PageMargins,
    copy_count=PrintEnginePropertyKey.PPK_CopyCount,
    supports_multiple_copies=PrintEnginePropertyKey.PPK_SupportsMultipleCopies,
    qpagesize=PrintEnginePropertyKey.PPK_QPageSize,
    qpagemargins=PrintEnginePropertyKey.PPK_QPageMargins,
    qpagelayout=PrintEnginePropertyKey.PPK_QPageLayout,
    custom_base=PrintEnginePropertyKey.PPK_CustomBase,
)


PrintEnginePropertyKeyStr = Literal["none", "auto", "long_side", "short_side"]


class PrintEngine(printsupport.QPrintEngine):
    """Defines an interface for how QPrinter interacts with a given printing subsystem."""

    def get_printer_state(self) -> PrintEnginePropertyKeyStr:
        return printsupport.printer.PRINTER_STATE.inverse[self.printerState()]



from __future__ import annotations

from prettyqt.docs import build_root, build_index

import mknodes


def build(project: mknodes.Project):
    root_nav = build_root.build_root(project)
    page = root_nav.add_index_page("Home", hide_nav=True)
    build_index.build_index(page)


# from prettyqt import prettyqtmarkdown, widgets
# app = widgets.app()
# print(root_nav)
# table = prettyqtmarkdown.MarkdownWidget()
# table.set_markdown(root_nav)
# table.show()
# app.exec()



from __future__ import annotations

import sys
from typing import Self

from prettyqt.qt import QtCore
from prettyqt.utils import datatypes, get_repr


class VersionNumber(QtCore.QVersionNumber):
    """Contains a version number with an arbitrary number of segments."""

    def __init__(self, *args, **kwargs):
        match args:
            case (str(),):
                args = [int(i) for i in args[0].split(".")][:3]
            case (tuple(),):
                args = args[0]
            # PySide6 Workaround:
            case (QtCore.QVersionNumber(),):
                args = (
                    args[0].majorVersion(),
                    args[0].minorVersion(),
                    args[0].microVersion(),
                )
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return get_repr(self, self.major(), self.minor(), self.micro())

    @property
    def _majorVersion(self) -> int:
        return self.majorVersion()

    @property
    def _minorVersion(self) -> int:
        return self.minorVersion()

    @property
    def _microVersion(self) -> int:
        return self.microVersion()

    __match_args__ = ("_majorVersion", "_minorVersion", "_microVersion")

    def __reduce__(self):
        return type(self), (self.major(), self.minor(), self.micro())

    def __str__(self):
        return self.toString()

    def __eq__(self, other):
        if isinstance(other, str | tuple):
            other = VersionNumber(other)
        return super().__eq__(other)

    def __hash__(self):
        return hash((self.major(), self.minor(), self.micro()))

    def __gt__(self, other: datatypes.SemanticVersionType):
        if isinstance(other, str | tuple):
            other = VersionNumber(other)
        return super().__gt__(other)

    def __ge__(self, other: datatypes.SemanticVersionType):
        if isinstance(other, str | tuple):
            other = VersionNumber(other)
        return super().__ge__(other)

    def __lt__(self, other: datatypes.SemanticVersionType):
        if isinstance(other, str | tuple):
            other = VersionNumber(other)
        return super().__lt__(other)

    def __le__(self, other: datatypes.SemanticVersionType):
        if isinstance(other, str | tuple):
            other = VersionNumber(other)
        return super().__le__(other)

    @classmethod
    def from_string(cls, version_string: str) -> Self:
        return cls(cls.fromString(version_string))

    @classmethod
    def get_qt_version(cls) -> Self:
        return cls(*[int(i) for i in QtCore.__version__.split(".")])

    @classmethod
    def get_python_version(cls) -> Self:
        return cls(*sys.version_info[:3])

    def major(self) -> int:
        """An integer representing the major version."""
        return self.majorVersion()

    def minor(self) -> int:
        """An integer representing the minor version."""
        return self.minorVersion()

    def micro(self) -> int:
        """An integer representing the micro version."""
        return self.microVersion()


if __name__ == "__main__":
    version = VersionNumber(3, 0, 0)



from __future__ import annotations

from prettyqt import core


class SignalMapper(core.ObjectMixin, core.QSignalMapper):
    """Bundles signals from identifiable senders."""

    def __getitem__(self, index: int | str | core.QObject) -> core.QObject:
        return self.mapping(index)

    def __delitem__(self, index: core.QObject):
        return self.removeMappings(index)

    def __setitem__(self, index: core.QObject, value: int | str | core.QObject):
        self.setMapping(index, value)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.Widget()
    mapper = SignalMapper()
    mapper[widget] = 1
    del mapper[widget]



"""Tests for `prettyqt` package."""

import sys

import pytest

import prettyqt
from prettyqt import core, qthelp
from prettyqt.qt import QtHelp


# from prettyqt.utils import InvalidParamError


def test_helplink():
    qtlink = QtHelp.QHelpLink()
    link = qthelp.HelpLink(qtlink)
    link.get_url()


def test_helpcontentitem():
    engine = qthelp.HelpEngine("")
    model = engine.get_content_model()
    repr(model)
    # item = model.get_content_item_at(core.QModelIndex())
    # item.get_url()
    # assert len(item) == 0


def test_helpcontentmodel():
    engine = qthelp.HelpEngine("")
    model = engine.get_content_model()
    repr(model)
    # model.get_content_item_at(core.QModelIndex())


def test_helpfilterdata():
    data = qthelp.HelpFilterData(QtHelp.QHelpFilterData())
    versions = [core.VersionNumber(1, 0, 0)]
    data.set_versions(versions)
    assert data.get_versions() == versions


def test_helpsearchresultwidget():
    core_engine = qthelp.HelpEngineCore("")
    engine = qthelp.HelpSearchEngine(core_engine)
    widget = engine.get_result_widget()
    widget.get_link_at(core.Point(1, 1))


@pytest.mark.skipif(
    sys.platform == "linux" and prettyqt.qt.API.startswith("pyside"),
    reason="Segmentation fault",
)
def test_helpsearchquerywidget():
    widget = qthelp.HelpSearchQueryWidget()
    assert widget is not None


def test_helpfilterengine():
    core_engine = qthelp.HelpEngineCore("")
    engine = core_engine.get_filter_engine()
    engine.get_available_versions()


def test_helpengine():
    engine = qthelp.HelpEngine("")
    engine.get_file_data(core.Url(""))
    engine.get_files("", "")


def test_helpsearchengine():
    core_engine = qthelp.HelpEngineCore("")
    engine = qthelp.HelpSearchEngine(core_engine)
    engine.search_results(0, 0)


def test_helpsearchresult():
    result = qthelp.HelpSearchResult(core.Url(""), "", "")
    result.get_url()



"""Tests for `prettyqt` package."""

from prettyqt import gui, test, widgets


def test_buttondelegate(qtbot):
    model = gui.StandardItemModel()
    test.AbstractItemModelTester(model)


def test_signalspy(qtbot):
    widget = widgets.Widget()
    test.SignalSpy(widget.windowIconChanged)
